(*
 * Fowl - the Free/Open Widget Library
 * Copyright (C) 2016 Kostas Michalopoulos
 *)
unit FObjects;
{$MODE OBJFPC}{$H+}
interface

const
  { Fowl version }
  FowlVersion = $00001000;
  FowlVersionString = 'Fowl version 0.1';

  { Fowl window class name }
  FowlWindowClassName = 'FowlWindow';

  { Stream modes }
  stCreate = $3C00;
  stOpenRead = $3D00;
  stOpenWrite = $3D01;
  stOpen = $3D02;

  { Steam status }
  stOk = 0;
  stError = -1;
  stInitError = -2;
  stReadError = -3;
  stWriteError = -4;
  stGetError = -5;
  stPutError = -6;

  { Object types }
  ot_First = $1000; // First available object type ID

  { Collection errors }
  coIndexError = -1;

  { Menu item type }
  it_Normal = 0;
  it_SubMenu = 1;
  it_Separator = 2;
  it_Check = 3;
  it_Radio = 4;

  { Window object flags }
  wb_NoFlags = $00;
  wb_KeyboardHandler = $01;
  wb_FromResource = $02;
  wb_AutoCreate = $04;
  wb_MDIChild = $08;
  wb_Transfer = $10;
  wb_HasMenu = $20;

  { Window error conditions }
  em_Ok = 0;
  em_InvalidWindow = -1;
  em_OutOfMemory = -2;
  em_InvalidClient = -3;
  em_InvalidChild = -4;
  em_InvalidMainWindow = -5;

  { Window show commands }
  sw_Hide = 0;
  sw_Show = 1;
  sw_ShowMaximized = 2;
  sw_ShowMinimized = 3;
  sw_ShowNormal = 4;

  { Transfer flags }
  tf_SizeData = 0;
  tf_GetData = 1;
  tf_SetData = 2;

  { Control notifications }
  cn_Clicked = 1;
  cn_Changed = 2;

  { Id numbers }
  id_None = -1;

type
  { Pointers to objects }
  PObject = ^TObject;
  PStream = ^TStream;
  PMemStream = ^TMemStream;
  PFileStream = ^TFileStream;
  PStreamRec = ^TStreamRec;
  PCollection = ^TCollection;
  PDC = ^TDC;
  PPaintStruct = ^TPaintStruct;
  PMenuItem = ^TMenuItem;
  PMenu = ^TMenu;
  PWindowsObject = ^TWindowsObject;
  PWindow = ^TWindow;
  PControl = ^TControl;
  PButton = ^TButton;
  PScrollBarTransferRec = ^TScrollBarTransferRec;
  PScrollBar = ^TScrollBar;
  PStatic = ^TStatic;
  PEdit = ^TEdit;
  PListBoxTransferRec = ^TListBoxTransferRec;
  PListBox = ^TListBox;
  PComboBoxTransferRec = ^TComboBoxTransferRec;
  PComboBox = ^TComboBox;
  PGroupBox = ^TGroupBox;
  PCheckBoxState = ^TCheckBoxState;
  PCheckBox = ^TCheckBox;
  PRadioButton = ^TRadioButton;
  PMDIClient = PWindowsObject;
  PApplication = ^TApplication;

  { TObject - base object for the rest of Fowl's objects }
  TObject = object
    constructor Init;
    procedure Free;
    destructor Done; virtual;
  end;

  { TStream - base object for streams }
  TStream = object(TObject)
    ErrorInfo: Int64;
    Status: Integer;
    constructor Init;
    procedure CopyFrom(var S: TStream; Count: Int64);
    procedure Error(Code: Integer; Info: Int64); virtual;
    function HasError: Boolean;
    procedure Flush; virtual;
    function Get: PObject;
    function GetPos: Int64; virtual;
    function GetSize: Int64; virtual;
    procedure Put(P: PObject);
    procedure Read(out Buf; Count: Int64); virtual;
    function ReadStr: string;
    procedure Reset;
    procedure Seek(Pos: Int64); virtual;
    procedure Truncate; virtual;
    procedure Write(const Buf; Count: Int64); virtual;
    procedure WriteStr(S: string);
  end;

  { TMemStream - memory backed stream }
  TMemStream = object(TStream)
    Buffer: Pointer;
    Size: Int64;
    Position: Int64;
    FreeBuffer: Boolean;
    constructor Init(InitialSize: Int64=0);
    constructor Wrap(DataBuffer: Pointer; DataSize: Int64);
    destructor Done; virtual;
    procedure SetSize(NewSize: Int64);
    function GetPos: Int64; virtual;
    function GetSize: Int64; virtual;
    procedure Read(out Buf; Count: Int64); virtual;
    procedure Seek(Pos: Int64); virtual;
    procedure Truncate; virtual;
    procedure Write(const Buf; Count: Int64); virtual;
  end;

  { TFileStream - file access stream }
  TFileStream = object(TStream)
    Handle: THandle;
    constructor Init(FileName: string; Mode: Integer);
    destructor Done; virtual;
    function GetPos: Int64; virtual;
    function GetSize: Int64; virtual;
    procedure Read(out Buf; Count: Int64); virtual;
    procedure Seek(Pos: Int64); virtual;
    procedure Truncate; virtual;
    procedure Write(const Buf; Count: Int64); virtual;
  end;

  { Stream error procedure type }
  TStreamError = procedure(var S: TStream);

  { TStreamRec - streamable object registry entry }
  TStreamRec = packed record
    ObjType: Cardinal;
    VmtLink: Pointer;
    Load: Pointer;
    Store: Pointer;
    Next: PStreamRec;
  end;

  { Collection test function type }
  TCollectionTest = function(Item: Pointer): Boolean;

  { Collection action procedure type }
  TCollectionAction = procedure(Item: Pointer);

  { TCollection - base type for item collections }
  TCollection = object(TObject)
    Count: Integer;
    Delta: Integer;
    Items: array of Pointer;
    constructor Init(ALimit: Integer=0; ADelta: Integer=16);
    constructor Load(var S: TStream);
    destructor Done; virtual;
    procedure AtDelete(Index: Integer);
    procedure AtFree(Index: Integer);
    procedure AtInsert(Index: Integer; Item: Pointer);
    procedure AtPut(Index: Integer; Item: Pointer);
    function At(Index: Integer): Pointer;
    procedure Delete(Item: Pointer);
    procedure DeleteAll;
    procedure Error(Code, Info: Integer); virtual;
    function FirstThat(Test: TCollectionTest): Pointer;
    procedure ForEach(Action: TCollectionAction);
    procedure FreeAll;
    procedure FreeItem(Item: Pointer); virtual;
    procedure Free(Item: Pointer);
    function GetItem(var S: TStream): Pointer; virtual;
    function IndexOf(Item: Pointer): Integer; virtual;
    procedure Insert(Item: Pointer); virtual;
    function LastThat(Test: TCollectionTest): Pointer;
    procedure Pack;
    procedure PutItem(var S: TStream; Item: Pointer); virtual;
    procedure SetLimit(ALimit: Integer); virtual;
    procedure Store(var S: TStream);
  end;

  { Collection error procedure type }
  TCollectionError = procedure(var C: TCollection; Error, Info: Integer);

  { TDC - device context for drawing operations }
  TDC = object(TObject)
    HDC: THandle;
  end;

  { TPaintStruct - information about the paint event }
  TPaintStruct = record
    DC: PDC;
    Erase: Boolean;
    X, Y, W, H: Integer;
  end;

  { TMenuItem - single menu item }
  TMenuItem = object(TObject)
    Title: string;
    Id: Integer;
    Flag: Integer;
    SubMenu: PMenu;
    constructor Init(ATitle: string; AnId: Integer; ItemFlag: Integer; ASubMenu: PMenu);
    constructor Load(var S: TStream);
    destructor Done; virtual;
    procedure SetEnabled(Enabled: Boolean);
    function GetEnabled: Boolean;
    procedure Store(var S: TStream);
  end;

  { TMenu - window system menu }
  TMenu = object(TObject)
    Items: PCollection;
    constructor Init;
    constructor Load(var S: TStream);
    destructor Done; virtual;
    function AppendMenu(ATitle: string; AnId: Integer; ItemFlag: Integer; ASubMenu: PMenu): PMenuItem; virtual;
    procedure AppendSeparator; virtual;
    procedure Store(var S: TStream);
  end;

  { TMessage - window system message }
  TMessage = record
    Receiver: THandle;
    Message: Integer;
    WParam: Int64;
    LParam: Int64;
    Result: Int64;
  end;

  { Window system object test function type }
  TWindowsObjectTest = function(P: PWindowsObject): Boolean;

  { Window system object action procedure type }
  TWindowsObjectAction = procedure(P: PWindowsObject);

  { TWindowsObject - base type for window system objects }
  TWindowsObject = object(TObject)
    ChildList: PWindowsObject;
    SiblingList: PWindowsObject;
    Flags: Byte;
    HWindow: THandle;
    Parent: PWindowsObject;
    Status: Integer;
    TransferBuffer: Pointer;
    constructor Init(AParent: PWindowsObject);
    constructor Load(var S: TStream);
    destructor Done; virtual;
    function CanClose: Boolean; virtual;
    function ChildWithId(Id: Integer): PWindowsObject; virtual;
    function ChildWithHandle(Handle: THandle): PWindowsObject;
    procedure CloseWindow;
    function Create: Boolean; virtual;
    function CreateChildren: Boolean;
    procedure DefChildProc(var Msg: TMessage); virtual;
    procedure DefCommandProc(var Msg: TMessage); virtual;
    procedure DefWndProc(var Msg: TMessage); virtual;
    procedure Destroy; virtual;
    procedure DisableAutoCreate;
    procedure DisableTransfer;
    procedure DispatchScroll(var Msg: TMessage); virtual;
    procedure EnableAutoCreate;
    procedure EnableKBHandler;
    procedure EnableTransfer;
    function FirstThat(Test: TWindowsObjectTest): PWindowsObject;
    procedure ForEach(Action: TWindowsObjectAction);
    procedure GetChildPtr(var S: TStream; out P: PWindowsObject);
    procedure GetChildren(var S: TStream);
    function GetClassName: string; virtual;
    function GetClient: PMDIClient; virtual;
    function GetId: Integer; virtual;
    procedure GetSiblingPtr(var S: TStream; out P: PWindowsObject);
    procedure GetWindowClass(var AWndClass); virtual;
    function IsFlagSet(Mask: Byte): Boolean;
    function Next: PWindowsObject;
    function Previous: PWindowsObject;
    procedure PutChildPtr(var S: TStream; P: PWindowsObject);
    procedure PutChildren(var S: TStream);
    procedure PutSiblingPtr(var S: TStream; P: PWindowsObject);
    function Register: Boolean; virtual;
    procedure SetFlags(Mask: Byte; OnOff: Boolean);
    procedure UpdateMenu; virtual;
    procedure SetupWindow; virtual;
    function CheckHandle: Boolean;
    procedure Show(ShowCmd: Integer); virtual;
    procedure Move(X, Y: Integer); virtual;
    procedure Resize(W, H: Integer); virtual;
    procedure Place(X, Y, W, H: Integer); virtual;
    procedure GetRectArea(out X, Y, W, H: Integer); virtual;
    procedure GetClientArea(out X, Y, W, H: Integer); virtual;
    procedure SetText(Text: string); virtual;
    function GetText: string; virtual;
    procedure Invalidate; virtual;
    procedure InvalidateArea(X, Y, W, H: Integer); virtual;
    procedure Repaint; virtual;
    procedure Enable(Flag: Boolean); virtual;
    function IsEnabled: Boolean; virtual;
    procedure Store(var S: TStream);
    function Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer; virtual;
    procedure TransferData(Direction: Integer); virtual;
    procedure ProcessMessage(var Msg: TMessage); virtual;
    function ProcessParentNotificationMessage(var Msg: TMessage): Boolean; virtual;
    function ProcessNotification(Child: PWindowsObject; NotifyCode: Integer): Boolean; virtual;
    function ProcessCommand(Command: Integer): Boolean; virtual;
  end;

  { TWindowAttr - attributes for a window }
  TWindowAttr = record
    Title: string;
    Style: Longint;
    ExStyle: Longint;
    X, Y, W, H: Integer;
    Param: Pointer;
    case Integer of
      0: (Menu: PMenu);
      1: (Id: Integer);
  end;

  { TWindow - base type for all windows and controls }
  TWindow = object(TWindowsObject)
    Attr: TWindowAttr;
    FocusChildHandle: PWindowsObject;
    constructor Init(AParent: PWindowsObject; ATitle: string);
    constructor Load(var S: TStream);
    destructor Done; virtual;
    function Create: Boolean; virtual;
    procedure DefWndProc(var Msg: TMessage); virtual;
    function GetId: Integer; virtual;
    procedure GetWindowClass(var AWndClass); virtual;
    procedure MoveToCenter;
    procedure Paint(var PaintDC: TDC; const PaintInfo: TPaintStruct); virtual;
    procedure SetMenu(AMenu: PMenu); virtual;
    function GetMenu: PMenu; virtual;
    procedure RemoveMenu; virtual;
    procedure UpdateMenu; virtual;
    procedure SetupWindow; virtual;
    procedure Move(X, Y: Integer); virtual;
    procedure Resize(W, H: Integer); virtual;
    procedure Place(X, Y, W, H: Integer); virtual;
    procedure GetRectArea(out X, Y, W, H: Integer); virtual;
    procedure SetText(Text: string); virtual;
    function GetText: string; virtual;
    function StoreText: Boolean; virtual;
    procedure Store(var S: TStream);
    procedure ProcessMessage(var Msg: TMessage); virtual;
  end;

  { TControl - base type for controls }
  TControl = object(TWindow)
    constructor Init(AParent: PWindowsObject; AnId: Integer; ATitle: string; X, Y, W, H: Integer);
    function GetClassName: string; virtual;
    function Register: Boolean; virtual;
    procedure ProcessMessage(var Msg: TMessage); virtual;
  end;

  { TButton - push button control }
  TButton = object(TControl)
    constructor Init(AParent: PWindowsObject; AnId: Integer; AText: string; X, Y, W, H: Integer; IsDefault: Boolean);
    function GetClassName: string; virtual;
    function ProcessParentNotificationMessage(var Msg: TMessage): Boolean; virtual;
  end;

  { Scroll bar data transfer record }
  TScrollBarTransferRec = packed record
    LowValue: Integer;
    HighValue: Integer;
    Position: Integer;
  end;

  { TScrollBar - horizontal or vertical scrollbar control }
  TScrollBar = object(TControl)
    LineMagnitude: Integer;
    PageMagnitude: Integer;
    constructor Init(AParent: PWindowsObject; AnId, X, Y, W, H: Integer; IsHScrollBar: Boolean);
    constructor Load(var S: TStream);
    function DeltaPos(Delta: Integer): Integer; virtual;
    function GetClassName: string; virtual;
    function GetPosition: Integer; virtual;
    procedure GetRange(out LoVal, HiVal: Integer); virtual;
    procedure SetPosition(ThumbPos: Integer); virtual;
    procedure SetRange(LoVal, HiVal: Integer); virtual;
    procedure SetupWindow; virtual;
    procedure Store(var S: TStream);
    function Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer; virtual;
    procedure ProcessSBMessage(var Msg: TMessage);
  end;

  { TStatic - static text control }
  TStatic = object(TControl)
    constructor Init(AParent: PWindowsObject; AnId: Integer; AText: string; X, Y, W, H: Integer);
    procedure Clear; virtual;
    function GetClassName: string; virtual;
    function Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer; virtual;
  end;

  { Text editor options }
  TEditOption = (eoMultiline, eoNoBorder, eoWordWrap, eoReadOnly, eoPassword);
  TEditOptions = set of TEditOption;

  { TEdit - text editor control }
  TEdit = object(TStatic)
    constructor Init(AParent: PWindowsObject; AnId: Integer; AText: string; X, Y, W, H: Integer; Options: TEditOptions);
    function CanUndo: Boolean; virtual;
    procedure ClearModify; virtual;
    procedure Copy; virtual;
    procedure Cut; virtual;
    function DeleteLine(LineNumber: Integer): Boolean; virtual;
    function DeleteSelection: Boolean; virtual;
    function DeleteSubText(StartPos, EndPos: Integer): Boolean; virtual;
    function GetClassName: string; virtual;
    function GetLine(LineNumber: Integer): string; virtual;
    function GetLineFromPos(CharPos: Integer): Integer; virtual;
    function GetLineIndex(LineNumber: Integer): Integer; virtual;
    function GetLineLength(LineNumber: Integer): Integer; virtual;
    function GetNumLines: Integer; virtual;
    function GetSelection(out StartPos, EndPos: Integer): Boolean; virtual;
    function GetSubText(StartPos, EndPos: Integer): string; virtual;
    procedure Insert(ATextString: string); virtual;
    function IsModified: Boolean; virtual;
    procedure Paste; virtual;
    procedure Scroll(HorizontalUnit, VerticalUnit: Integer); virtual;
    function Search(StartPos: Integer; AText: string; CaseSensitive: Boolean): Integer; virtual;
    function SetSelection(StartPos, EndPos: Integer): Boolean; virtual;
    procedure Undo; virtual;
    procedure SetupWindow; virtual;
    function StoreText: Boolean; virtual;
    function ProcessParentNotificationMessage(var Msg: TMessage): Boolean; virtual;
  end;

  { List box data transfer record }
  TListBoxTransferRec = packed record
    Strings: array of string;
    Selected: array of Boolean;
    SelIndex: Integer;
  end;

  { List box options }
  TListBoxOption = (loMultiSelect, loNoBorder, loSortItems);
  TListBoxOptions = set of TListBoxOption;

  { TListBox - list box control }
  TListBox = object(TControl)
    constructor Init(AParent: PWindowsObject; AnId, X, Y, W, H: Integer; Options: TListBoxOptions);
    function AddString(AString: string): Integer; virtual;
    procedure ClearList; virtual;
    function DeleteString(Index: Integer): Integer; virtual;
    function GetClassName: string; virtual;
    function GetCount: Integer; virtual;
    function GetSelIndex: Integer; virtual;
    function IsSelected(Index: Integer): Boolean; virtual;
    function GetSelString: string; virtual;
    function GetString(Index: Integer): string; virtual;
    function InsertString(AString: string; Index: Integer): Integer; virtual;
    function SetSelIndex(Index: Integer): Integer; virtual;
    function SetSelected(Index: Integer; Selected: Boolean): Boolean; virtual;
    function SetSelString(AString: string; Index: Integer): Integer; virtual;
    procedure SetupWindow; virtual;
    function StoreText: Boolean; virtual;
    function Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer; virtual;
    function ProcessParentNotificationMessage(var Msg: TMessage): Boolean; virtual;
  end;

  { Combo box data transfer record }
  TComboBoxTransferRec = packed record
    Strings: array of string;
    SelIndex: Integer;
    Text: string;
  end;

  { Combo box options }
  TComboBoxOption = (coNoBorder, coSortItems, coDropDownList);
  TComboBoxOptions = set of TComboBoxOption;

  { TComboBox - combo box control }
  TComboBox = object(TControl)
    constructor Init(AParent: PWindowsObject; AnId, X, Y, W, H: Integer; Options: TComboBoxOptions);
    function AddString(AString: string): Integer; virtual;
    procedure ClearList; virtual;
    function DeleteString(Index: Integer): Integer; virtual;
    function GetClassName: string; virtual;
    function GetCount: Integer; virtual;
    function GetSelIndex: Integer; virtual;
    function GetSelString: string; virtual;
    function GetString(Index: Integer): string; virtual;
    function InsertString(AString: string; Index: Integer): Integer; virtual;
    function SetSelIndex(Index: Integer): Integer; virtual;
    function SetSelString(AString: string; Index: Integer): Integer; virtual;
    procedure HideList; virtual;
    procedure ShowList; virtual;
    procedure SetupWindow; virtual;
    function Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer; virtual;
    function ProcessParentNotificationMessage(var Msg: TMessage): Boolean; virtual;
  end;

  { TGroupBox - group box control }
  TGroupBox = object(TControl)
    NotifyParent: Boolean;
    constructor Init(AParent: PWindowsObject; AnId: Integer; ATitle: string; X, Y, W, H: Integer);
    constructor Load(var S: TStream);
    function GetClassName: string; virtual;
    procedure SelectionChanged(ControlId: Integer); virtual;
    procedure SetupWindow; virtual;
    procedure Store(var S: TStream);
    function ProcessNotification(Child: PWindowsObject; NotifyCode: Integer): Boolean; virtual;
  end;

  { Check box state }
  TCheckBoxState = (csUnchecked, csChecked, csGrayed);

  { Check box options }
  TCheckBoxOption = (coAutoCheck, co3State);
  TCheckBoxOptions = set of TCheckBoxOption;

  { TCheckBox - check box control }
  TCheckBox = object(TControl)
    constructor Init(AParent: PWindowsObject; AnId: Integer; ATitle: string; X, Y, W, H: Integer; Options: TCheckBoxOptions);
    procedure Check; virtual;
    function GetClassName: string; virtual;
    function GetCheck: TCheckBoxState; virtual;
    procedure SetCheck(CheckFlag: TCheckBoxState); virtual;
    procedure Toggle; virtual;
    function Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer; virtual;
    procedure Uncheck; virtual;
    function ProcessParentNotificationMessage(var Msg: TMessage): Boolean; virtual;
  end;

  { TRadioButton - radio button control }
  TRadioButton = object(TCheckBox)
    constructor Init(AParent: PWindowsObject; AnId: Integer; ATitle: string; X, Y, W, H: Integer);
    procedure SetCheck(CheckFlag: TCheckBoxState); virtual;
    function ProcessParentNotificationMessage(var Msg: TMessage): Boolean; virtual;
  end;

  { TApplication - base type for the application singleton }
  TApplication = object(TObject)
    KBHandlerWnd: PWindowsObject;
    MainWindow: PWindowsObject;
    Name: string;
    Status: Integer;
    constructor Init(AName: string);
    destructor Done; virtual;
    function CanClose: Boolean; virtual;
    procedure Error(ErrorCode: Integer); virtual;
    function ExecDialog(ADialog: PWindowsObject): Integer; virtual;
    procedure ShowMessage(Text: string); virtual;
    procedure InitApplication; virtual;
    procedure InitInstance; virtual;
    procedure InitMainWindow; virtual;
    function MakeWindow(AWindowsObject: PWindowsObject): PWindowsObject; virtual;
    procedure MessageLoop; virtual;
    function ProcessAccels(var Message: TMessage): Boolean; virtual;
    function ProcessAppMsg(var Message: TMessage): Boolean; virtual;
    function ProcessDlgMsg(var Message: TMessage): Boolean; virtual;
    function ProcessMDIAccels(var Message: TMessage): Boolean; virtual;
    procedure Run; virtual;
    procedure Terminate; virtual;
    procedure SetKBHandler(AWindowsObject: PWindowsObject);
    function ValidWindow(AWindowsObject: PWindowsObject): PWindowsObject;
  end;

var
  StreamError: TStreamError;            // Stream error handler
  StreamVersion: Cardinal;              // Library version to use for streaming
  CollectionError: TCollectionError;    // Collection error handler
  Application: PApplication;            // Application singleton

{ Called to raise Runtime Error 211 }
procedure Abstract;
{ Called to register a new streamable object type }
procedure RegisterType(var S: TStreamRec);

implementation

uses
  Windows;
  
{$IFDEF CPUAMD64}
const
  GWLP_USERDATA = -21;
  GCLP_WNDPROC = -24;
{$ENDIF}

var
  StreamRecHead: PStreamRec;
  ButtonSubclassInstalled: Boolean;
  OriginalButtonProc: WndProc;

{ Local }
procedure DefaultStreamError(var S: TStream);

  procedure ShowMsg(Msg: string);
  begin
    if Assigned(Application) then
      Application^.ShowMessage(Msg)
    else
      MessageBoxA(0, PAnsiChar(Msg), 'Stream Error', MB_OK);
  end;

begin
  case S.Status of
    stError: ShowMsg('Stream Access Error');
    stInitError: ShowMsg('Stream Initialization Error');
    stReadError: ShowMsg('Stream Read Error');
    stWriteError: ShowMsg('Stream Write Error');
    stGetError: ShowMsg('Stream Object Read Error');
    stPutError: ShowMsg('Stream Object Write Error');
  end;
end;

function FindType(ObjType: Cardinal): PStreamRec;
var
  Entry: PStreamRec;
begin
  Entry:=StreamRecHead;
  while Assigned(Entry) do begin
    if Entry^.ObjType=ObjType then Exit(Entry);
    Entry:=Entry^.Next;
  end;
  Result:=nil;
end;

function FindStreamRec(P: PObject): PStreamRec;
var
  Entry: PStreamRec;
  VmtLink: Pointer;
begin
  VmtLink:=PPointer(P)^;
  Entry:=StreamRecHead;
  while Assigned(Entry) do begin
    if Entry^.VmtLink=VmtLink then Exit(Entry);
    Entry:=Entry^.Next;
  end;
  Result:=nil;
end;

function FOB_WindowProc(Window: HWND; AMessage: UINT; WParam: WPARAM;
  LParam: LPARAM): LRESULT; stdcall;
var
  Obj: PWindowsObject;
  Msg: TMessage;
  LPCS: PCreateStruct;
begin
  if AMessage=wm_Create then begin
    LPCS:=PCreateStruct(LParam);
    Obj:=LPCS^.lpCreateParams;
    if Assigned(Obj) then begin
      // Set HWindow now because the call to SetupWindow can be
      // made before the CreateWindow call returns
      Obj^.HWindow:=Window;
      {$IFDEF CPUAMD64}
      SetWindowLongPtr(Window, GWLP_USERDATA, LONG_PTR(Obj));
      {$ELSE}
      SetWindowLong(Window, GWL_USERDATA, LONG(Obj));
      {$ENDIF}
    end;
  end else
    {$IFDEF CPUAMD64}
    Obj:=PWindowsObject(GetWindowLongPtr(Window, GWLP_USERDATA));
    {$ELSE}
    Obj:=PWindowsObject(GetWindowLong(Window, GWL_USERDATA));
    {$ENDIF}
  if Assigned(Obj) then begin
    Msg.Receiver:=Window;
    Msg.Message:=AMessage;
    Msg.WParam:=WParam;
    Msg.LParam:=LParam;
    Msg.Result:=0;
    Obj^.ProcessMessage(Msg);
    Result:=Msg.Result;
  end else begin
    Result:=DefWindowProc(Window, AMessage, WParam, LParam);
  end;
end;

function ReplacementButtonProc(Window: HWND; AMessage: UINT; WParam: WPARAM;
  LParam: LPARAM): LRESULT; stdcall;
var
  Style: Integer;
  TopHandle: THandle;
  Top, Child: PWindowsObject;
  Group: PGroupBox;
  ClassName: array [0..Length(FowlWindowClassName)] of AnsiChar;
  Msg: TMessage;
begin
  // Check if we need to forward the notification message from a
  // group box to its parent
  if (AMessage=wm_Command) and (LParam <> 0) then begin
    Style:=GetWindowLong(Window, GWL_STYLE);
    if (Style and bs_GroupBox)=bs_GroupBox then begin
      // Make sure this is a window inside a Fowl window
      TopHandle:=Window;
      while GetParent(TopHandle) <> 0 do TopHandle:=GetParent(TopHandle);
      GetClassName(TopHandle, @ClassName[0], Length(FowlWindowClassName) + 1);
      if ClassName=FowlWindowClassName then begin
        {$IFDEF CPUAMD64}
        Top:=PWindowsObject(GetWindowLongPtr(TopHandle, GWLP_USERDATA));
        {$ELSE}
        Top:=PWindowsObject(GetWindowLong(TopHandle, GWL_USERDATA));
        {$ENDIF}
        if Assigned(Top) then begin
          Group:=PGroupBox(Top^.ChildWithHandle(Window));
          if Assigned(Group) and
             Group^.NotifyParent and
             Assigned(Group^.Parent) then begin
            Child:=Group^.ChildWithHandle(LParam);
            if Assigned(Child) then begin
              Msg.Receiver:=Window;
              Msg.Message:=AMessage;
              Msg.WParam:=WParam;
              Msg.LParam:=LParam;
              Msg.Result:=0;
              Child^.ProcessParentNotificationMessage(Msg);
            end;
          end;
        end;
      end;
    end;
  end;
  Result:=CallWindowProc(OriginalButtonProc, Window, AMessage, WParam, LParam);
end;

{ Global }
procedure Abstract;
begin
  RunError(211);
end;

procedure RegisterType(var S: TStreamRec);
begin
  S.Next:=StreamRecHead;
  StreamRecHead:=@S;
end;

{ TObject }
constructor TObject.Init;
begin
end;

procedure TObject.Free;
var
  This: PObject;
begin
  This:=@Self;
  if This <> nil then begin
    Dispose(This, Done);
  end;
end;

destructor TObject.Done;
begin
end;

{ TStream }
constructor TStream.Init;
begin
  inherited Init;
  ErrorInfo:=0;
  Status:=stOk;
end;

procedure TStream.CopyFrom(var S: TStream; Count: Int64);
const
  BufferSize = 65536; // Microsoft recommends 64KB buffers for I/O
var
  Buffer: Pointer;
  HaveRead, ToRead: Int64;
begin
  if HasError or S.HasError then Exit;
  GetMem(Buffer, BufferSize);
  HaveRead:=0;
  while HaveRead < Count do begin
    ToRead:=BufferSize;
    if HaveRead + ToRead > Count then ToRead:=Count - HaveRead;
    S.Read(Buffer^, ToRead);
    if S.HasError then Error(stWriteError, stReadError);
    if HasError then Break;
    Write(Buffer^, ToRead);
    if HasError then Break;
  end;
  FreeMem(Buffer, BufferSize);
end;

procedure TStream.Error(Code: Integer; Info: Int64);
begin
  // Disallow multiple errors to pile up
  if HasError then Exit;
  ErrorInfo:=Info;
  Status:=Code;
  if Assigned(StreamError) then StreamError(Self);
end;

function TStream.HasError: Boolean;
begin
  Result:=Status <> stOk;
end;

procedure TStream.Flush;
begin
  Abstract;
end;

function TStream.Get: PObject;
type
  TLoadCtor = function(Obj, VMT, Stream: Pointer): Pointer;
var
  ObjType: Cardinal;
  StreamRec: PStreamRec;
  Marker: Byte;
begin
  if HasError then Exit(nil);
  Read(Marker, 1);
  if Marker=0 then Exit(nil);
  Read(ObjType, 4);
  if HasError then Exit(nil);
  StreamRec:=FindType(ObjType);
  if (not Assigned(StreamRec)) or
     (not Assigned(StreamRec^.Load)) then begin
    Error(stGetError, ObjType);
    Exit(nil);
  end;
  Result:=TLoadCtor(StreamRec^.Load)(nil, StreamRec^.VmtLink, @Self);
end;

function TStream.GetPos: Int64;
begin
  Abstract;
  Result:=0;
end;

function TStream.GetSize: Int64;
begin
  Abstract;
  Result:=0;
end;

procedure TStream.Put(P: PObject);
type
  TStoreMethod = procedure(Obj, Stream: Pointer);
var
  StreamRec: PStreamRec;
  Marker: Byte;
begin
  if HasError then Exit;
  if not Assigned(P) then begin
    Marker:=0;
    Write(Marker, 1);
    Exit;
  end;
  StreamRec:=FindStreamRec(P);
  if (not Assigned(StreamRec)) or
     (not Assigned(StreamRec^.Store)) then begin
    Error(stPutError, 0);
    Exit;
  end;
  Marker:=1;
  Write(Marker, 1);
  Write(StreamRec^.ObjType, 4);
  if not HasError then TStoreMethod(StreamRec^.Store)(P, @Self);
end;

procedure TStream.Read(out Buf; Count: Int64);
begin
  Abstract;
end;

function TStream.ReadStr: string;
var
  Len: Cardinal;
  {$IFNDEF VER2_0}
  S: Utf8String;
  {$ENDIF}
begin
  if HasError then Exit('');
  Read(Len, 4);
  if HasError then Exit('');
  {$IFNDEF VER2_0}
  SetLength(S, Len);
  if Len=0 then Exit('');
  Read((@S[1])^, Len);
  Result:=S;
  {$ELSE}
  SetLength(Result, Len);
  if Len=0 then Exit('');
  Read((@Result[1])^, Len);
  {$ENDIF}
end;

procedure TStream.Reset;
begin
  Status:=stOk;
  ErrorInfo:=0;
end;

procedure TStream.Seek(Pos: Int64);
begin
  Abstract;
end;

procedure TStream.Truncate;
begin
  Abstract;
end;

procedure TStream.Write(const Buf; Count: Int64);
begin
  Abstract;
end;

procedure TStream.WriteStr(S: string);
var
  Len: Cardinal;
  {$IFNDEF VER2_0}
  SS: Utf8String;
  {$ENDIF}
begin
  if HasError then Exit;
  {$IFNDEF VER2_0}
  SS:=S;
  Len:=Length(SS);
  {$ELSE}
  Len:=Length(S);
  {$ENDIF}
  Write(Len, 4);
  if Len > 0 then begin
    {$IFDEF VER2_0}
    Write((@SS[1])^, Len);
    {$ELSE}
    Write((@S[1])^, Len);
    {$ENDIF}
  end;
end;

{ TMemStream }
constructor TMemStream.Init(InitialSize: Int64);
begin
  inherited Init;
  if InitialSize > 0 then Buffer:=GetMem(InitialSize) else Buffer:=nil;
  Size:=InitialSize;
  if Size < 0 then Size:=0;
  Position:=0;
  FreeBuffer:=True;
end;

constructor TMemStream.Wrap(DataBuffer: Pointer; DataSize: Int64);
begin
  Buffer:=DataBuffer;
  Size:=DataSize;
  Position:=0;
  FreeBuffer:=False;
end;

destructor TMemStream.Done;
begin
  if FreeBuffer then FreeMem(Buffer);
  inherited Done;
end;

procedure TMemStream.SetSize(NewSize: Int64);
begin
  if HasError then Exit;
  if NewSize=Size then Exit;
  Buffer:=ReAllocMem(Buffer, NewSize);
  Size:=NewSize;
end;

function TMemStream.GetPos: Int64;
begin
  Result:=Position;
end;

function TMemStream.GetSize: Int64;
begin
  Result:=Size;
end;

procedure TMemStream.Read(out Buf; Count: Int64);
begin
  if HasError or (Count=0) then Exit;
  if Position + Count > Size then begin
    Error(stReadError, 0);
    Exit;
  end;
  // TODO: split the move in multiples when moving >2G in 32bit systems
  Move((Buffer + Position)^, Buf, Count);
  Inc(Position, Count);
end;

procedure TMemStream.Seek(Pos: Int64);
begin
  if HasError then Exit;
  if Pos < 0 then begin
    Error(stError, 0);
    Exit;
  end;
  Position:=Pos;
end;

procedure TMemStream.Truncate;
begin
  if HasError then Exit;
  SetSize(Position);
end;

procedure TMemStream.Write(const Buf; Count: Int64);
begin
  if HasError or (Count=0) then Exit;
  if Position + Count > Size then SetSize(Position + Count);
  if HasError then Exit;
  Move(Buf, (Buffer + Position)^, Count);
  Inc(Position, Count);
end;

{ TFileStream }
constructor TFileStream.Init(FileName: string; Mode: Integer);
begin
  inherited Init;
  case Mode of
    stCreate:
      Handle:=CreateFileA(PAnsiChar(FileName),
        GENERIC_READ or GENERIC_WRITE, 0, nil,
        CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
    stOpenRead:
      Handle:=CreateFileA(PAnsiChar(FileName),
        GENERIC_READ, 0, nil,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    stOpenWrite:
      Handle:=CreateFileA(PAnsiChar(FileName),
        GENERIC_WRITE, 0, nil,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    stOpen:
      Handle:=CreateFileA(PAnsiChar(FileName),
        GENERIC_READ or GENERIC_WRITE, 0, nil,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    else Handle:=THandle(-1);
  end;
  if Handle=THandle(-1) then begin
    Error(stInitError, 0);
    Exit;
  end;
end;

destructor TFileStream.Done;
begin
  if Handle <> THandle(-1) then CloseHandle(Handle);
  inherited Done;
end;

function TFileStream.GetPos: Int64;
var
  LoPart, HiPart: DWORD;
begin
  if HasError or (Handle=THandle(-1)) then Exit(0);
  LoPart:=SetFilePointer(Handle, 0, @HiPart, FILE_CURRENT);
  if LoPart=$FFFFFFFF then begin
    Error(stError, 0);
    Exit(0);
  end;
  Result:=Int64(LoPart) + (Int64(HiPart) shl 32);
end;

function TFileStream.GetSize: Int64;
var
  Here: Int64;
  LoPart, HiPart: DWORD;
begin
  if HasError or (Handle=THandle(-1)) then Exit(0);
  Here:=GetPos;
  if HasError then Exit(0);
  LoPart:=SetFilePointer(Handle, 0, @HiPart, FILE_END);
  if LoPart=$FFFFFFFF then begin
    Error(stError, 0);
    Exit(0);
  end;
  Result:=Int64(LoPart) + (Int64(HiPart) shl 32);
  Seek(Here);
end;

procedure TFileStream.Read(out Buf; Count: Int64);
var
  Got: DWORD;
begin
  if HasError or (Handle=THandle(-1)) then Exit;
  // TODO: handle >2G reads
  if not ReadFile(Handle, Buf, Count, Got, nil) then begin
    Error(stReadError, 0);
    Exit;
  end;
  if Got <> Count then begin
    Error(stReadError, 0);
    Exit;
  end;
end;

procedure TFileStream.Seek(Pos: Int64);
var
  LoPart, HiPart: DWORD;
begin
  if HasError or (Handle=THandle(-1)) then Exit;
  if Pos < 0 then begin
    Error(stError, 0);
    Exit;
  end;
  LoPart:=Pos and $FFFFFFFF;
  HiPart:=Pos shr 32;
  SetFilePointer(Handle, LoPart, @HiPart, FILE_BEGIN);
  if GetPos <> Pos then begin
    Error(stError, 0);
    Exit;
  end;
end;

procedure TFileStream.Truncate;
begin
  if HasError or (Handle=THandle(-1)) then Exit;
  if not SetEndOfFile(Handle) then Error(stError, 0);
end;

procedure TFileStream.Write(const Buf; Count: Int64);
var
  Wrote: DWORD;
begin
  if HasError or (Handle=THandle(-1)) then Exit;
  // TODO: handle >2G writes
  if not WriteFile(Handle, Buf, Count, Wrote, nil) then begin
    Error(stWriteError, 0);
    Exit;
  end;
  if Wrote <> Count then begin
    Error(stWriteError, 0);
    Exit;
  end;
end;

{ TCollection }
constructor TCollection.Init(ALimit: Integer; ADelta: Integer);
begin
  inherited Init;
  SetLength(Items, ALimit);
  Delta:=ADelta;
  Count:=0;
end;

constructor TCollection.Load(var S: TStream);
var
  I: Integer;
begin
  inherited Init;
  Items:=nil;
  Delta:=16;
  Count:=0;
  if not S.HasError then begin
    S.Read(Count, 4);
    if S.HasError then Exit;
    if Count < 0 then Exit;
    SetLength(Items, Count);
    for I:=0 to Count - 1 do begin
      Items[I]:=GetItem(S);
      if S.HasError then Break;
    end;
  end;
end;

destructor TCollection.Done;
begin
  Items:=nil;
  inherited Done;
end;

procedure TCollection.AtDelete(Index: Integer);
var
  I: Integer;
begin
  if (Index < 0) or (Index >= Count) then begin
    Error(coIndexError, Index);
    Exit;
  end;
  for I:=Index to Count - 1 do Items[I]:=Items[I + 1];
  Dec(Count);
end;

procedure TCollection.AtFree(Index: Integer);
var
  I: Integer;
begin
  if (Index < 0) or (Index >= Count) then begin
    Error(coIndexError, Index);
    Exit;
  end;
  FreeItem(Items[Index]);
  for I:=Index to Count - 1 do Items[I]:=Items[I + 1];
  Dec(Count);
end;

procedure TCollection.AtInsert(Index: Integer; Item: Pointer);
var
  I: Integer;
begin
  if (Index < 0) or (Index > Count) then begin
    Error(coIndexError, Index);
    Exit;
  end;
  if Count=Length(Items) then SetLimit(Length(Items) + Delta);
  Inc(Count);
  for I:=Count - 1 downto Index + 1 do
    Items[I]:=Items[I - 1];
  Items[Index]:=Item;
end;

procedure TCollection.AtPut(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index >= Count) then begin
    Error(coIndexError, Index);
    Exit;
  end;
  Items[Index]:=Item;
end;

function TCollection.At(Index: Integer): Pointer;
begin
  if (Index < 0) or (Index >= Count) then begin
    Error(coIndexError, Index);
    Exit(nil);
  end;
  Result:=Items[Index];
end;

procedure TCollection.Delete(Item: Pointer);
begin
  AtDelete(IndexOf(Item));
end;

procedure TCollection.DeleteAll;
begin
  Count:=0;
end;

procedure TCollection.Error(Code, Info: Integer);
begin
  if Assigned(CollectionError) then
    CollectionError(Self, Code, Info)
  else
    RunError(213);
end;

function TCollection.FirstThat(Test: TCollectionTest): Pointer;
var
  I: Integer;
begin
  if not Assigned(Test) then Exit(nil);
  for I:=0 to Count - 1 do
    if Test(Items[I]) then Exit(Items[I]);
  Result:=nil;
end;

procedure TCollection.ForEach(Action: TCollectionAction);
var
  I: Integer;
begin
  if not Assigned(Action) then Exit;
  for I:=0 to Count - 1 do Action(Items[I]);
end;

procedure TCollection.FreeAll;
var
  I, C: Integer;
begin
  C:=Count;
  Count:=0;
  for I:=0 to C - 1 do FreeItem(Items[I]);
end;

procedure TCollection.FreeItem(Item: Pointer);
begin
  if Item <> nil then Dispose(PObject(Item), Done);
end;

procedure TCollection.Free(Item: Pointer);
begin
  Delete(Item);
  FreeItem(Item);
end;

function TCollection.GetItem(var S: TStream): Pointer;
begin
  Result:=S.Get;
end;

function TCollection.IndexOf(Item: Pointer): Integer;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    if Item=Items[I] then Exit(I);
  Result:=-1;
end;

procedure TCollection.Insert(Item: Pointer);
begin
  AtInsert(Count, Item);
end;

function TCollection.LastThat(Test: TCollectionTest): Pointer;
var
  I: Integer;
begin
  if not Assigned(Test) then Exit(nil);
  for I:=Count - 1 downto 0 do
    if Test(Items[I]) then Exit(Items[I]);
  Result:=nil;
end;

procedure TCollection.Pack;
var
  I: Integer;
begin
  for I:=Count - 1 downto 0 do
    if Items[I]=nil then AtDelete(I);
end;

procedure TCollection.PutItem(var S: TStream; Item: Pointer);
begin
  S.Put(Item);
end;

procedure TCollection.SetLimit(ALimit: Integer);
begin
  if ALimit < Count then ALimit:=Count;
  SetLength(Items, ALimit);
end;

procedure TCollection.Store(var S: TStream);
var
  I: Integer;
begin
  if S.HasError then Exit;
  S.Write(Count, 4);
  for I:=0 to Count - 1 do begin
    PutItem(S, Items[I]);
    if S.HasError then Break;
  end;
end;

{ TMenuItem }
constructor TMenuItem.Init(ATitle: string; AnId: Integer; ItemFlag: Integer; ASubMenu: PMenu);
begin
  inherited Init;
  Title:=ATitle;
  Id:=AnId;
  Flag:=ItemFlag;
  if ItemFlag=it_SubMenu then begin
    Id:=id_None;
    SubMenu:=ASubMenu
  end else begin
    if Assigned(ASubMenu) then
      Dispose(ASubMenu, Done);
    SubMenu:=nil;
  end;
end;

constructor TMenuItem.Load(var S: TStream);
begin
  inherited Init;
  Title:=S.ReadStr;
  S.Read(Id, 4);
  S.Read(Flag, 4);
  SubMenu:=PMenu(S.Get);
end;

destructor TMenuItem.Done;
begin
  if Assigned(SubMenu) then Dispose(SubMenu, Done);
  inherited Done;
end;

procedure TMenuItem.SetEnabled(Enabled: Boolean);
begin
  if Enabled then
    Flag:=Flag and (not $00010000)
  else
    Flag:=Flag or $00010000;
end;

function TMenuItem.GetEnabled: Boolean;
begin
  Result:=(Flag and $00010000)=$00010000;
end;

procedure TMenuItem.Store(var S: TStream);
begin
  S.WriteStr(Title);
  S.Write(Id, 4);
  S.Write(Flag, 4);
  S.Put(SubMenu);
end;

{ TMenu }
constructor TMenu.Init;
begin
  inherited Init;
  New(Items, Init);
end;

constructor TMenu.Load(var S: TStream);
begin
  inherited Init;
  Items:=PCollection(S.Get);
  if not Assigned(Items) then New(Items, Init);
end;

destructor TMenu.Done;
begin
  Dispose(Items, Done);
  inherited Done;
end;

function TMenu.AppendMenu(ATitle: string; AnId: Integer; ItemFlag: Integer; ASubMenu: PMenu): PMenuItem;
begin
  New(Result, Init(ATitle, AnId, ItemFlag, ASubMenu));
  Items^.Insert(Result);
end;

procedure TMenu.AppendSeparator;
begin
  AppendMenu('', id_None, it_Separator, nil);
end;

procedure TMenu.Store(var S: TStream);
begin
  S.Put(Items);
end;

{ TWindowsObject }
constructor TWindowsObject.Init(AParent: PWindowsObject);
begin
  inherited Init;
  ChildList:=nil;
  Flags:=0;
  HWindow:=0;
  Parent:=AParent;
  Status:=0;
  TransferBuffer:=nil;
  EnableAutoCreate;
  if Assigned(Parent) then begin
    SiblingList:=Parent^.ChildList;
    Parent^.ChildList:=@Self;
  end else
    SiblingList:=nil;
end;

constructor TWindowsObject.Load(var S: TStream);
begin
  inherited Init;
  HWindow:=0;
  Parent:=nil;
  TransferBuffer:=nil;
  S.Read(Status, 4);
  S.Read(Flags, 1);
  GetChildren(S);
end;

destructor TWindowsObject.Done;
begin
  if HWindow <> 0 then Destroy;
  if Application^.MainWindow=@Self then Application^.MainWindow:=nil;
  if Application^.KBHandlerWnd=@Self then Application^.KBHandlerWnd:=nil;
end;

function TWindowsObject.CanClose: Boolean;
var
  Child: PWindowsObject;
begin
  Child:=ChildList;
  while Assigned(Child) do begin
    if not Child^.CanClose then Exit(False);
    Child:=Child^.Next;
  end;
  Result:=True;
end;

function TWindowsObject.ChildWithId(Id: Integer): PWindowsObject;
var
  Child: PWindowsObject;
begin
  if Id=GetId then Exit(@Self);
  Child:=ChildList;
  while Assigned(Child) do begin
    Result:=Child^.ChildWithId(Id);
    if Assigned(Result) then Exit;
    Child:=Child^.Next;
  end;
  Result:=nil;
end;

function TWindowsObject.ChildWithHandle(Handle: THandle): PWindowsObject;
var
  Child: PWindowsObject;
begin
  if Handle=0 then Exit(nil);
  if Handle=HWindow then Exit(@Self);
  Child:=ChildList;
  while Assigned(Child) do begin
    Result:=Child^.ChildWithHandle(Handle);
    if Assigned(Result) then Exit;
    Child:=Child^.Next;
  end;
  Result:=nil;
end;

procedure TWindowsObject.CloseWindow;
begin
  if CanClose then Destroy;
end;

function TWindowsObject.Create: Boolean;
begin
  Abstract;
  Result:=False;
end;

function TWindowsObject.CreateChildren: Boolean;
var
  Child: PWindowsObject;
begin
  Child:=ChildList;
  while Assigned(Child) do begin
    if Child^.IsFlagSet(wb_AutoCreate) then begin
      if not Child^.Create then Status:=em_InvalidChild;
    end;
    Child:=Child^.Next;
  end;
  Result:=True;
end;

procedure TWindowsObject.DefChildProc(var Msg: TMessage);
begin
  Msg.Result:=0;
end;

procedure TWindowsObject.DefCommandProc(var Msg: TMessage);
begin
  Msg.Result:=0;
end;

procedure TWindowsObject.DefWndProc(var Msg: TMessage);
begin
end;

procedure TWindowsObject.Destroy;
var
  Child: PWindowsObject;
begin
  if HWindow <> 0 then DestroyWindow(HWindow);
  Child:=ChildList;
  while Assigned(Child) do begin
    Child^.EnableAutoCreate;
    Child:=Child^.Next;
  end;
end;

procedure TWindowsObject.DisableAutoCreate;
begin
  SetFlags(wb_AutoCreate, False);
end;

procedure TWindowsObject.DisableTransfer;
begin
  SetFlags(wb_Transfer, False);
end;

procedure TWindowsObject.DispatchScroll(var Msg: TMessage);
var
  SB: PScrollBar;
begin
  if Msg.LParam <> 0 then begin
    SB:=PScrollBar(ChildWithHandle(THandle(Msg.LParam)));
    if Assigned(SB) then SB^.ProcessSBMessage(Msg);
  end;
end;

procedure TWindowsObject.EnableAutoCreate;
begin
  SetFlags(wb_AutoCreate, True);
end;

procedure TWindowsObject.EnableKBHandler;
begin
  SetFlags(wb_KeyboardHandler, True);
end;

procedure TWindowsObject.EnableTransfer;
begin
  SetFlags(wb_Transfer, True);
end;

function TWindowsObject.FirstThat(Test: TWindowsObjectTest): PWindowsObject;
var
  Child: PWindowsObject;
begin
  if not Assigned(Test) then Exit(nil);
  Child:=ChildList;
  while Assigned(Child) do begin
    if Test(Child) then Exit(Child);
    Child:=Child^.Next;
  end;
  Result:=nil;
end;

procedure TWindowsObject.ForEach(Action: TWindowsObjectAction);
var
  Child: PWindowsObject;
begin
  if not Assigned(Action) then Exit;
  Child:=ChildList;
  while Assigned(Child) do begin
    Action(Child);
    Child:=Child^.Next;
  end;
end;

procedure TWindowsObject.GetChildPtr(var S: TStream; out P: PWindowsObject);
begin
  P:=PWindowsObject(S.Get);
  if Assigned(P) then begin
    P^.SiblingList:=ChildList;
    ChildList:=P;
    P^.Parent:=@Self;
  end;
end;

procedure TWindowsObject.GetChildren(var S: TStream);
var
  Child: PWindowsObject;
  I, ChildCount: Integer;
begin
  S.Read(ChildCount, 4);
  if not S.HasError then begin
    if ChildCount > 0 then begin
      GetChildPtr(S, Child);
      if Assigned(Child) then begin
        for I:=1 to ChildCount - 1 do begin
          Child^.GetSiblingPtr(S, Child);
          if not Assigned(Child) then Break;
        end
      end;
    end;
  end;
end;

function TWindowsObject.GetClassName: string;
begin
  Result:=FowlWindowClassName;
end;

function TWindowsObject.GetClient: PMDIClient;
begin
  Result:=nil;
end;

function TWindowsObject.GetId: Integer;
begin
  Result:=id_None;
end;

procedure TWindowsObject.GetSiblingPtr(var S: TStream; out P: PWindowsObject);
begin
  if S.HasError then begin
    P:=nil;
    SiblingList:=nil;
  end else begin
    P:=PWindowsObject(S.Get);
    if Assigned(P) then begin
      P^.Parent:=Parent;
      P^.SiblingList:=@Self;
      Parent^.ChildList:=P;
    end;
  end;
end;

procedure TWindowsObject.GetWindowClass(var AWndClass);
begin
end;

function TWindowsObject.IsFlagSet(Mask: Byte): Boolean;
begin
  Result:=(Flags and Mask)=Mask;
end;

function TWindowsObject.Next: PWindowsObject;
begin
  Result:=SiblingList;
end;

function TWindowsObject.Previous: PWindowsObject;
var
  Child: PWindowsObject;
begin
  if (Parent=nil) or (Parent^.ChildList=@Self) then Exit(nil);
  Child:=Parent^.ChildList;
  while Assigned(Child) do begin
    if Child^.Next=@Self then Exit(Child);
    Child:=Child^.Next;
  end;
  Result:=nil; // Shouldn't happen
end;

procedure TWindowsObject.PutChildPtr(var S: TStream; P: PWindowsObject);
begin
  S.Put(P);
end;

procedure TWindowsObject.PutChildren(var S: TStream);
var
  Child: PWindowsObject;
  ChildCount: Integer;
begin
  if S.HasError then Exit;
  ChildCount:=0;
  Child:=ChildList;
  while Assigned(Child) do begin
    Inc(ChildCount);
    Child:=Child^.Next;
  end;
  S.Write(ChildCount, 4);
  if S.HasError then Exit;
  Child:=ChildList;
  if Assigned(Child) then begin
    PutChildPtr(S, Child);
    if S.HasError then Exit;
    while Assigned(Child^.Next) do begin
      Child^.PutSiblingPtr(S, Child^.Next);
      if S.HasError then Exit;
      Child:=Child^.Next;
    end;
  end;
end;

procedure TWindowsObject.PutSiblingPtr(var S: TStream; P: PWindowsObject);
begin
  S.Put(P);
end;

function TWindowsObject.Register: Boolean;
var
  WC: TWndClass;
begin
  // Check if the class already exists
  FillChar(WC, SizeOf(WC), 0);
  if GetClassInfo(HInstance, PAnsiChar(AnsiString(GetClassName)), WC) then Exit(True);
  FillChar(WC, SizeOf(WC), 0);
  WC.lpfnWndProc:=@FOB_WindowProc;
  WC.hInstance:=HInstance;
  WC.hIcon:=LoadIcon(0, IDI_APPLICATION);
  WC.hCursor:=LoadCursor(0, IDC_ARROW);
  WC.hbrBackground:=COLOR_BTNFACE + 1;
  WC.lpszClassName:=PAnsiChar(AnsiString(GetClassName));
  GetWindowClass(WC);
  Result:=RegisterClassA(@WC) <> 0;
end;

procedure TWindowsObject.SetFlags(Mask: Byte; OnOff: Boolean);
begin
  if OnOff then Flags:=Flags or Mask else Flags:=Flags and (not Mask);
end;

procedure TWindowsObject.UpdateMenu;
begin
end;

procedure TWindowsObject.SetupWindow;
begin
  UpdateMenu;
  CreateChildren;
  TransferData(tf_SetData);
end;

function TWindowsObject.CheckHandle: Boolean;
begin
  if HWindow=0 then begin
    Result:=False;
    Status:=em_InvalidWindow;
  end else
    Result:=True;
end;

procedure TWindowsObject.Show(ShowCmd: Integer);
begin
  if not CheckHandle then Exit;
  case ShowCmd of
    sw_Hide: ShowWindow(HWindow, Windows.SW_HIDE);
    sw_Show: ShowWindow(HWindow, Windows.SW_SHOW);
    sw_ShowMaximized: ShowWindow(HWindow, Windows.SW_SHOWMAXIMIZED);
    sw_ShowMinimized: ShowWindow(HWindow, Windows.SW_SHOWMINIMIZED);
    sw_ShowNormal: ShowWindow(HWindow, Windows.SW_SHOWNORMAL);
  end;
end;

procedure TWindowsObject.Move(X, Y: Integer);
begin
  if not CheckHandle then Exit;
  SetWindowPos(HWindow, 0, X, Y, 0, 0,
    swp_NoActivate or
    swp_NoReposition or
    swp_NoSize or
    swp_NoZOrder);
end;

procedure TWindowsObject.Resize(W, H: Integer);
begin
  if not CheckHandle then Exit;
  SetWindowPos(HWindow, 0, 0, 0, W, H,
    swp_NoActivate or
    swp_NoReposition or
    swp_NoMove or
    swp_NoZOrder);
end;

procedure TWindowsObject.Place(X, Y, W, H: Integer);
begin
  if not CheckHandle then Exit;
  SetWindowPos(HWindow, 0, X, Y, W, H,
    swp_NoActivate or
    swp_NoReposition or
    swp_NoZOrder);
end;

procedure TWindowsObject.GetRectArea(out X, Y, W, H: Integer);
var
  R: TRect;
begin
  if not CheckHandle then Exit;
  GetWindowRect(HWindow, R);
  X:=R.left;
  Y:=R.top;
  W:=R.right - R.left;
  H:=R.bottom - R.top;
end;

procedure TWindowsObject.GetClientArea(out X, Y, W, H: Integer);
var
  R: TRect;
begin
  if not CheckHandle then Exit;
  GetClientRect(HWindow, R);
  X:=R.left;
  Y:=R.top;
  W:=R.right - R.left;
  H:=R.bottom - R.top;
end;

procedure TWindowsObject.SetText(Text: string);
begin
  if not CheckHandle then Exit;
  SetWindowTextA(HWindow, PAnsiChar(Text));
end;

function TWindowsObject.GetText: string;
var
  Len: Integer;
  WT: PAnsiChar;
begin
  if not CheckHandle then Exit('');
  Len:=GetWindowTextLength(HWindow) + 1;
  WT:=GetMem(Len);
  GetWindowTextA(HWindow, WT, Len);
  WT[Len]:=#0;
  Result:=WT;
  FreeMem(WT);
end;

procedure TWindowsObject.Invalidate;
begin
  if not CheckHandle then Exit;
  InvalidateRect(HWindow, nil, False);
end;

procedure TWindowsObject.InvalidateArea(X, Y, W, H: Integer);
var
  R: TRect;
begin
  if not CheckHandle then Exit;
  R.left:=X;
  R.top:=Y;
  R.right:=X + W;
  R.bottom:=Y + H;
  InvalidateRect(HWindow, R, False);
end;

procedure TWindowsObject.Repaint;
begin
  if not CheckHandle then Exit;
  RedrawWindow(HWindow, nil, 0, rdw_Erase or rdw_InternalPaint);
end;

procedure TWindowsObject.Enable(Flag: Boolean);
begin
  if not CheckHandle then Exit;
  EnableWindow(HWindow, Flag);
end;

function TWindowsObject.IsEnabled: Boolean;
begin
  if not CheckHandle then Exit(False);
  Result:=IsWindowEnabled(HWindow);
end;

procedure TWindowsObject.Store(var S: TStream);
begin
  S.Write(Status, 4);
  S.Write(Flags, 1);
  PutChildren(S);
end;

function TWindowsObject.Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer;
begin
  Result:=0;
end;

procedure TWindowsObject.TransferData(Direction: Integer);
begin
  if (not IsFlagSet(wb_Transfer)) or (TransferBuffer=nil) then Exit;
  if (Direction <> tf_SetData) and (Direction <> tf_GetData) then Exit;
  Transfer(TransferBuffer, Direction);
end;

procedure TWindowsObject.ProcessMessage(var Msg: TMessage);
var
  Child: PWindowsObject;
begin
  case Msg.Message of
    wm_Activate:
      if IsFlagSet(wb_KeyboardHandler) then 
        if Msg.WParam in [wa_Active, wa_ClickActive] then begin
          Application^.SetKBHandler(@Self);
        end;
    wm_Close:
      if Application^.MainWindow=@Self then begin
        if Application^.CanClose then Destroy;
      end else begin
        if CanClose then Destroy;
      end;
    wm_Command:
      if Msg.LParam in [0, 1] then begin
        if not ProcessCommand(LoWord(Msg.WParam)) then
          DefCommandProc(Msg);
      end else begin
        Child:=ChildWithHandle(THandle(Msg.LParam));
        if Assigned(Child) then begin
          if not Child^.ProcessParentNotificationMessage(Msg) then
            DefChildProc(Msg);
        end else
          DefChildProc(Msg);
      end;
    wm_Destroy:
      if Application^.MainWindow=@Self then Application^.Terminate;
    wm_HScroll, wm_VScroll:
      DispatchScroll(Msg);
    wm_NCDestroy:
      HWindow:=0;
    wm_QueryEndSession:
      if (Application^.MainWindow=@Self) and Application^.CanClose then
        Msg.Result:=1
      else
        Msg.Result:=0;
    else
      DefWndProc(Msg);
  end;
end;

function TWindowsObject.ProcessParentNotificationMessage(var Msg: TMessage): Boolean;
begin
  Result:=False;
end;

function TWindowsObject.ProcessNotification(Child: PWindowsObject; NotifyCode: Integer): Boolean;
begin
  Result:=False;
end;

function TWindowsObject.ProcessCommand(Command: Integer): Boolean;
begin
  Result:=False;
end;

{ TWindow }
constructor TWindow.Init(AParent: PWindowsObject; ATitle: string);
begin
  inherited Init(AParent);
  FillChar(Attr, SizeOf(Attr), 0);
  FocusChildHandle:=nil;
  Attr.Title:=ATitle;
  Attr.Style:=ws_OverlappedWindow;
  Attr.X:=cw_UseDefault;
  Attr.Y:=cw_UseDefault;
  Attr.W:=cw_UseDefault;
  Attr.H:=cw_UseDefault;
  Attr.Param:=@Self;
end;

constructor TWindow.Load(var S: TStream);
begin
  inherited Load(S);
  FillChar(Attr, SizeOf(Attr), 0);
  FocusChildHandle:=nil;
  if StoreText then Attr.Title:=S.ReadStr;
  S.Read(Attr.Style, 4);
  S.Read(Attr.ExStyle, 4);
  S.Read(Attr.X, 4);
  S.Read(Attr.Y, 4);
  S.Read(Attr.W, 4);
  S.Read(Attr.H, 4);
  if S.HasError then Exit;
  if IsFlagSet(wb_HasMenu) then
    Attr.Menu:=PMenu(S.Get)
  else
    S.Read(Attr.Id, 4);
  Attr.Param:=@Self;
end;

destructor TWindow.Done;
begin
  if IsFlagSet(wb_HasMenu) then Dispose(Attr.Menu, Done);
  inherited Done;
end;

function TWindow.Create: Boolean;
var
  WndParent: THandle = 0;
begin
  if not Register then Exit(False);
  if Assigned(Parent) then WndParent:=Parent^.HWindow;
  HWindow:=CreateWindowExA(Attr.ExStyle, PAnsiChar(GetClassName),
    PAnsiChar(Attr.Title), Attr.Style, Attr.X, Attr.Y, Attr.W, Attr.H,
    WndParent, 0, HInstance, Attr.Param);
  if HWindow=0 then begin
    Status:=em_InvalidWindow;
    Exit(False);
  end;
  SetupWindow;
  Result:=True;
end;

procedure TWindow.DefWndProc(var Msg: TMessage);
begin
  Msg.Result:=DefWindowProc(Msg.Receiver, Msg.Message, Msg.WParam, Msg.LParam);
end;

function TWindow.GetId: Integer;
begin
  if (Attr.Style and ws_Child)=ws_Child then
    Result:=Attr.Id
  else
    Result:=id_None;
end;

procedure TWindow.GetWindowClass(var AWndClass);
var
  WC: TWndClass absolute AWndClass;
begin
  WC.Style:=cs_HRedraw or cs_VRedraw;
  WC.hIcon:=LoadIcon(0, IDI_APPLICATION);
  WC.hCursor:=LoadCursor(0, IDC_ARROW);
  WC.hbrBackground:=COLOR_BTNFACE + 1;
end;

procedure TWindow.MoveToCenter;
var
  PR: TRect;
begin
  if Parent=nil then
    GetWindowRect(GetDesktopWindow, PR)
  else begin
    Parent^.GetRectArea(PR.Left, PR.Top, PR.Right, PR.Bottom);
    PR.Right:=PR.Right + PR.Left;
    PR.Bottom:=PR.Bottom + PR.Top;
  end;
  Move(PR.Left + ((PR.Right - PR.Left) - Attr.W) div 2,
       PR.Top + ((PR.Bottom - PR.Top) - Attr.H) div 2);
end;

procedure TWindow.Paint(var PaintDC: TDC; const PaintInfo: TPaintStruct);
begin
end;

procedure TWindow.SetMenu(AMenu: PMenu);
begin
  if Assigned(Parent) then Exit;
  if IsFlagSet(wb_HasMenu) then RemoveMenu;
  SetFlags(wb_HasMenu, True);
  Attr.Menu:=AMenu;
  if HWindow <> 0 then UpdateMenu;
end;

function TWindow.GetMenu: PMenu;
begin
  if Assigned(Parent) then Exit(nil);
  Result:=Attr.Menu;
end;

procedure TWindow.RemoveMenu;
begin
  if Assigned(Parent) then Exit;
  if not IsFlagSet(wb_HasMenu) then Exit;
  SetFlags(wb_HasMenu, False);
  Dispose(Attr.Menu, Done);
  Attr.Menu:=nil;
  if HWindow <> 0 then UpdateMenu;
end;

procedure TWindow.UpdateMenu;
var
  NewMenu, OldMenu: THandle;

  function CreateMenuFor(Menu: PMenu): THandle;
  var
    I: Integer;
    Item: PMenuItem;
    MIT: TMenuItemInfoA;
  begin
    if not Assigned(Menu) then Exit(0);
    Result:=CreateMenu;
    for I:=0 to Menu^.Items^.Count - 1 do begin
      Item:=PMenuItem(Menu^.Items^.At(I));
      FillChar(MIT, SizeOf(MIT), 0);
      MIT.cbSize:=SizeOf(MIT);
      MIT.fMask:=miim_Type;
      case (Item^.Flag and $FFFF) of
        it_Normal: begin
          MIT.fMask:=MIT.fMask or miim_Id;
          MIT.fType:=mft_String;
          MIT.wId:=Item^.Id;
          MIT.dwTypeData:=PAnsiChar(Item^.Title);
        end;
        it_SubMenu: begin
          MIT.fMask:=MIT.fMask or miim_SubMenu;
          MIT.fType:=mft_String;
          MIT.hSubMenu:=CreateMenuFor(Item^.SubMenu);
          MIT.dwTypeData:=PAnsiChar(Item^.Title);
        end;
        it_Separator: MIT.fType:=mft_Separator;
        it_Check: begin
          MIT.fMask:=MIT.fMask or miim_Id;
          MIT.fType:=mft_String;
          MIT.fState:=mfs_Checked;
          MIT.wId:=Item^.Id;
          MIT.dwTypeData:=PAnsiChar(Item^.Title);
        end;
        it_Radio: begin
          MIT.fMask:=MIT.fMask or miim_Id;
          MIT.fType:=mft_String or mft_RadioCheck;
          MIT.fState:=mfs_Checked;
          MIT.wId:=Item^.Id;
          MIT.dwTypeData:=PAnsiChar(Item^.Title);
        end;
      end;
      if not Item^.GetEnabled then MIT.fState:=MIT.fState or mfs_Grayed;
      InsertMenuItem(Result, I, True, MIT);
    end;
  end;

begin
  if Assigned(Parent) then Exit;
  if not CheckHandle then Exit;
  OldMenu:=Windows.GetMenu(HWindow);
  if Assigned(Attr.Menu) then begin
    NewMenu:=CreateMenuFor(Attr.Menu);
  end else begin
    NewMenu:=0;
  end;
  Windows.SetMenu(HWindow, NewMenu);
  if OldMenu <> 0 then DestroyMenu(OldMenu);
end;

procedure TWindow.SetupWindow;
begin
  if HWindow <> 0 then SendMessage(HWindow, wm_SetFont, WParam(GetStockObject(Default_GUI_Font)), 1);
  inherited SetupWindow;
  // TODO: if Assigned(Scroller) then Scroller^.SetSBarRange;
end;

procedure TWindow.Move(X, Y: Integer);
begin
  if HWindow=0 then begin
    Attr.X:=X;
    Attr.Y:=Y;
  end else inherited Move(X, Y);
end;

procedure TWindow.Resize(W, H: Integer);
begin
  if HWindow=0 then begin
    Attr.W:=W;
    Attr.H:=H;
  end else inherited Resize(W, H);
end;

procedure TWindow.Place(X, Y, W, H: Integer);
begin
  if HWindow=0 then begin
    Attr.X:=X;
    Attr.Y:=Y;
    Attr.W:=W;
    Attr.H:=H;
  end else inherited Place(X, Y, W, H);
end;

procedure TWindow.GetRectArea(out X, Y, W, H: Integer);
begin
  if HWindow=0 then begin
    X:=Attr.X;
    Y:=Attr.Y;
    W:=Attr.W;
    H:=Attr.H;
  end else inherited GetRectArea(X, Y, W, H);
end;

procedure TWindow.SetText(Text: string);
begin
  Attr.Title:=Text;
  if HWindow <> 0 then begin
    inherited SetText(Text);
  end;
end;

function TWindow.GetText: string;
begin
  if HWindow <> 0 then begin
    Result:=inherited GetText;
    Attr.Title:=Result;
  end else
    Result:=Attr.Title;
end;

function TWindow.StoreText: Boolean;
begin
  Result:=True;
end;

procedure TWindow.Store(var S: TStream);
var
  Style, ExStyle: Integer;
begin
  inherited Store(S);
  if HWindow <> 0 then begin
    Style:=GetWindowLong(HWindow, GWL_STYLE);
    ExStyle:=GetWindowLong(HWindow, GWL_EXSTYLE);
  end else begin
    Style:=Attr.Style;
    ExStyle:=Attr.ExStyle;
  end;
  if StoreText then S.WriteStr(GetText);
  S.Write(Style, 4);
  S.Write(ExStyle, 4);
  S.Write(Attr.X, 4);
  S.Write(Attr.Y, 4);
  S.Write(Attr.W, 4);
  S.Write(Attr.H, 4);
  if IsFlagSet(wb_HasMenu) then
    S.Put(Attr.Menu)
  else
    S.Write(Attr.Id, 4);
end;

procedure TWindow.ProcessMessage(var Msg: TMessage);
var
  Placement: TWindowPlacement;
  PaintStruct: Windows.TPaintStruct;
  PaintInfo: TPaintStruct;
  UseUpdateRect, UseBeginPaint: Boolean;
  DC: HDC;
  FowlDC: TDC;
  R: TRect;
begin
  case Msg.Message of
    wm_Activate:
      if Msg.WParam in [wa_Active, wa_ClickActive] then
        FocusChildHandle:=ChildWithHandle(THandle(Msg.LParam));
    wm_HScroll: begin
      inherited ProcessMessage(Msg);
      Exit;
    end;
    wm_VScroll: begin
      inherited ProcessMessage(Msg);
      Exit;
    end;
    wm_Move: begin
      // Update the X/Y only if the window is in normal state
      GetWindowPlacement(Msg.Receiver, Placement);
      if not (Placement.showCmd in [Windows.sw_ShowMaximized,
                                    Windows.sw_ShowMinimized]) then begin
        GetWindowRect(Msg.Receiver, R);
        if GetParent(Msg.Receiver) <> 0 then begin
          ScreenToClient(GetParent(Msg.Receiver), PPoint(@R));
        end;
        Attr.X:=R.left;
        Attr.Y:=R.top;
      end;
    end;
    wm_Paint: begin
      DC:=HDC(Msg.WParam);
      UseUpdateRect:=GetUpdateRect(Msg.Receiver, R, False);
      UseBeginPaint:=UseUpdateRect;
      PaintInfo.DC:=@FowlDC;
      if UseBeginPaint then begin
        DC:=BeginPaint(Msg.Receiver, PaintStruct);
        if DC <> 0 then begin
          FowlDC.HDC:=DC;
          PaintInfo.Erase:=PaintStruct.fErase;
          PaintInfo.X:=PaintStruct.rcPaint.left;
          PaintInfo.Y:=PaintStruct.rcPaint.top;
          PaintInfo.W:=PaintStruct.rcPaint.right - PaintStruct.rcPaint.left;
          PaintInfo.H:=PaintStruct.rcPaint.bottom - PaintStruct.rcPaint.top;
          Paint(FowlDC, PaintInfo);
          EndPaint(Msg.Receiver, PaintStruct);
        end else begin
          UseBeginPaint:=False;
        end;
      end;
      if not UseBeginPaint then begin
        if not UseUpdateRect then GetClientRect(Msg.Receiver, R);
        PaintInfo.Erase:=False;
        PaintInfo.X:=R.left;
        PaintInfo.Y:=R.top;
        PaintInfo.W:=R.right - R.left;
        PaintInfo.H:=R.bottom - R.top;
        if DC=0 then begin
          DC:=GetDC(Msg.Receiver);
          if DC <> 0 then begin
            FowlDC.HDC:=DC;
            Paint(FowlDC, PaintInfo);
            ReleaseDC(Msg.Receiver, DC);
          end;
        end else begin
          FowlDC.HDC:=DC;
          Paint(FowlDC, PaintInfo);
        end;
      end;
    end;
    wm_Size: begin
      // Update the W/H only if the window is in normal state
      GetWindowPlacement(Msg.Receiver, Placement);
      if not (Placement.showCmd in [Windows.sw_ShowMaximized,
                                    Windows.sw_ShowMinimized]) then begin
        GetWindowRect(Msg.Receiver, R);
        Attr.W:=R.right - R.left;
        Attr.H:=R.bottom - R.top;
      end;
    end;
  end;
  inherited ProcessMessage(Msg);
end;

{ TControl }
constructor TControl.Init(AParent: PWindowsObject; AnId: Integer; ATitle: string; X, Y, W, H: Integer);
begin
  inherited Init(AParent, ATitle);
  Attr.Id:=AnId;
  Attr.X:=X;
  Attr.Y:=Y;
  Attr.W:=W;
  Attr.H:=H;
  Attr.Style:=ws_Child or ws_Visible or ws_Group or ws_TabStop;
end;

function TControl.GetClassName: string;
begin
  Abstract;
  Result:='';
end;

function TControl.Register: Boolean;
begin
  Result:=True;
end;

procedure TControl.ProcessMessage(var Msg: TMessage);
begin
  if Msg.Message=wm_Paint then
    DefWndProc(Msg)
  else
    inherited ProcessMessage(Msg);
end;

{ TButton }
constructor TButton.Init(AParent: PWindowsObject; AnId: Integer; AText: string; X, Y, W, H: Integer; IsDefault: Boolean);
begin
  inherited Init(AParent, AnId, AText, X, Y, W, H);
  if IsDefault then
    Attr.Style:=Attr.Style or bs_DefPushButton
  else
    Attr.Style:=Attr.Style or bs_PushButton;
end;

function TButton.GetClassName: string;
begin
  Result:='BUTTON';
end;

function TButton.ProcessParentNotificationMessage(var Msg: TMessage): Boolean;
begin
  Result:=False;
  case HiWord(Msg.WParam) of
    Windows.bn_Clicked:
      Result:=Parent^.ProcessNotification(@Self, cn_Clicked);
  end;
end;

{ TScrollBar }
constructor TScrollBar.Init(AParent: PWindowsObject; AnId, X, Y, W, H: Integer; IsHScrollBar: Boolean);
begin
  inherited Init(AParent, AnId, '', X, Y, W, H);
  if IsHScrollBar then
    Attr.Style:=Attr.Style or sbs_Horz
  else
    Attr.Style:=Attr.Style or sbs_Vert;
  LineMagnitude:=1;
  PageMagnitude:=10;
end;

constructor TScrollBar.Load(var S: TStream);
begin
  inherited Load(S);
  S.Read(LineMagnitude, 4);
  S.Read(PageMagnitude, 4);
end;

function TScrollBar.DeltaPos(Delta: Integer): Integer;
var
  LoVal, HiVal: Integer;
begin
  if not CheckHandle then Exit(0);
  GetScrollRange(HWindow, sb_Ctl, @LoVal, @HiVal);
  Result:=GetScrollPos(HWindow, sb_Ctl) + Delta;
  if Result < LoVal then Result:=LoVal;
  if Result > HiVal then Result:=HiVal;
  SetScrollPos(HWindow, sb_Ctl, Result, True);
end;

function TScrollBar.GetClassName: string;
begin
  Result:='SCROLLBAR';
end;

function TScrollBar.GetPosition: Integer;
begin
  if not CheckHandle then Exit(0);
  Result:=GetScrollPos(HWindow, sb_Ctl);
end;

procedure TScrollBar.GetRange(out LoVal, HiVal: Integer);
begin
  if not CheckHandle then Exit;
  GetScrollRange(HWindow, sb_Ctl, LoVal, HiVal);
end;

procedure TScrollBar.SetPosition(ThumbPos: Integer);
var
  LoVal, HiVal: Integer;
begin
  if not CheckHandle then Exit;
  GetScrollRange(HWindow, sb_Ctl, LoVal, HiVal);
  if ThumbPos < LoVal then ThumbPos:=LoVal;
  if ThumbPos > HiVal then ThumbPos:=HiVal;
  SetScrollPos(HWindow, sb_Ctl, ThumbPos, True);
end;

procedure TScrollBar.SetRange(LoVal, HiVal: Integer);
begin
  if not CheckHandle then Exit;
  SetScrollRange(HWindow, sb_Ctl, LoVal, HiVal, True);
end;

procedure TScrollBar.SetupWindow;
begin
  inherited SetupWindow;
  SetRange(0, 100);
  SetPosition(0);
end;

procedure TScrollBar.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(LineMagnitude, 4);
  S.Write(PageMagnitude, 4);
end;

function TScrollBar.Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer;
begin
  Result:=SizeOf(TScrollBarTransferRec);
  case TransferFlag of
    tf_GetData: begin
      GetRange(PScrollBarTransferRec(DataPtr)^.LowValue, PScrollBarTransferRec(DataPtr)^.HighValue);
      PScrollBarTransferRec(DataPtr)^.Position:=GetPosition;
    end;
    tf_SetData: begin
      SetRange(PScrollBarTransferRec(DataPtr)^.LowValue, PScrollBarTransferRec(DataPtr)^.HighValue);
      SetPosition(PScrollBarTransferRec(DataPtr)^.Position);
    end;
    tf_SizeData: begin end;
    else Result:=0;
  end;
end;

procedure TScrollBar.ProcessSBMessage(var Msg: TMessage);

  procedure FireChanged;
  begin
    if Assigned(Parent) then Parent^.ProcessNotification(@Self, cn_Changed);
  end;

begin
  case LoWord(Msg.WParam) of
    sb_Bottom: begin
      SetPosition($7FFFFFFF);
      FireChanged;
    end;
    sb_EndScroll: FireChanged;
    sb_LineLeft: begin
      DeltaPos(-LineMagnitude);
      FireChanged;
    end;
    sb_LineRight: begin
      DeltaPos(LineMagnitude);
      FireChanged;
    end;
    sb_PageLeft: begin
      DeltaPos(-PageMagnitude);
      FireChanged;
    end;
    sb_PageRight: begin
      DeltaPos(PageMagnitude);
      FireChanged;
    end;
    sb_ThumbPosition, sb_ThumbTrack: begin
      SetPosition(HiWord(Msg.WParam));
      FireChanged;
    end;
    sb_Top: begin
      SetPosition(0);
      FireChanged;
    end;
  end;
end;

{ TStatic }
constructor TStatic.Init(AParent: PWindowsObject; AnId: Integer; AText: string; X, Y, W, H: Integer);
begin
  inherited Init(AParent, AnId, AText, X, Y, W, H);
  Attr.Style:=Attr.Style and (not ws_TabStop);
  Attr.Style:=Attr.Style or ss_Left;
end;

procedure TStatic.Clear;
begin
  SetText('');
end;

function TStatic.GetClassName: string;
begin
  Result:='STATIC';
end;

function TStatic.Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer;
begin
  Result:=SizeOf(PString);
  case TransferFlag of
    tf_GetData: PString(DataPtr)^:=GetText;
    tf_SetData: SetText(PString(DataPtr)^);
    tf_SizeData: begin end;
    else Result:=0;
  end;
end;

{ TEdit }
constructor TEdit.Init(AParent: PWindowsObject; AnId: Integer; AText: string; X, Y, W, H: Integer; Options: TEditOptions);
begin
  inherited Init(AParent, AnId, AText, X, Y, W, H);
  Attr.Style:=Attr.Style or es_Left or es_AutoHScroll ;
  if not (eoNoBorder in Options) then Attr.ExStyle:=Attr.ExStyle or ws_ex_ClientEdge;
  if eoReadOnly in Options then Attr.Style:=Attr.Style or es_ReadOnly;
  if eoPassword in Options then Attr.Style:=Attr.Style or es_Password;
  if eoMultiline in Options then begin
    Attr.Style:=Attr.Style or es_Multiline or es_AutoVScroll or
      ws_HScroll or ws_VScroll;
    if eoWordWrap in Options then Attr.Style:=Attr.Style - es_AutoHScroll - ws_HScroll;
  end;
end;

function TEdit.CanUndo: Boolean;
begin
  if not CheckHandle then Exit(False);
  Result:=SendMessage(HWindow, em_CanUndo, 0, 0) <> 0;
end;

procedure TEdit.ClearModify;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, em_SetModify, 0, 0);
end;

procedure TEdit.Copy;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, wm_Copy, 0, 0);
end;

procedure TEdit.Cut;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, wm_Cut, 0, 0);
end;

function TEdit.DeleteLine(LineNumber: Integer): Boolean;
begin
  if not CheckHandle then Exit(False);
end;

function TEdit.DeleteSelection: Boolean;
begin
  if not CheckHandle then Exit(False);
  SendMessageA(HWindow, em_ReplaceSel, 1, LParam(PAnsiChar('')));
  Result:=True;
end;

function TEdit.DeleteSubText(StartPos, EndPos: Integer): Boolean;
begin
  if not CheckHandle then Exit(False);
  if not SetSelection(StartPos, EndPos) then Exit(False);
  Result:=DeleteSelection;
end;

function TEdit.GetClassName: string;
begin
  Result:='EDIT';
end;

function TEdit.GetLine(LineNumber: Integer): string;
var
  Len: Integer;
  Buff: PAnsiChar;
begin
  if not CheckHandle then Exit('');
  Len:=GetLineLength(LineNumber);
  Buff:=GetMem(Len);
  SendMessage(HWindow, em_GetLine, LineNumber, LParam(@Buff[0]));
  Buff[Len]:=#0;
  Result:=Buff;
  FreeMem(Buff);
end;

function TEdit.GetLineFromPos(CharPos: Integer): Integer;
begin
  if not CheckHandle then Exit(0);
  Result:=SendMessage(HWindow, em_LineFromChar, CharPos, 0);
end;

function TEdit.GetLineIndex(LineNumber: Integer): Integer;
begin
  if not CheckHandle then Exit(0);
  Result:=SendMessage(HWindow, em_LineIndex, LineNumber, 0);
end;

function TEdit.GetLineLength(LineNumber: Integer): Integer;
var
  Index: Integer;
begin
  if not CheckHandle then Exit(0);
  Index:=GetLineIndex(LineNumber);
  if Index=-1 then Exit(0);
  Result:=SendMessage(HWindow, em_LineLength, Index, 0);
end;

function TEdit.GetNumLines: Integer;
begin
  if not CheckHandle then Exit(0);
  Result:=SendMessage(HWindow, em_GetLineCount, 0, 0);
end;

function TEdit.GetSelection(out StartPos, EndPos: Integer): Boolean;
begin
  if not CheckHandle then Exit(False);
  SendMessage(HWindow, em_GetSel, WParam(@StartPos), WParam(@EndPos));
  Result:=True;
end;

function TEdit.GetSubText(StartPos, EndPos: Integer): string;
begin
  if not CheckHandle then Exit('');
  Result:=System.Copy(GetText, StartPos + 1, EndPos - StartPos);
end;

procedure TEdit.Insert(ATextString: string);
begin
  if not CheckHandle then Exit;
  SendMessageA(HWindow, em_ReplaceSel, 1, LParam(PAnsiChar(ATextString)));
end;

function TEdit.IsModified: Boolean;
begin
  if not CheckHandle then Exit(False);
  Result:=SendMessage(HWindow, em_GetModify, 0, 0) <> 0;
end;

procedure TEdit.Paste;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, wm_Paste, 0, 0);
end;

procedure TEdit.Scroll(HorizontalUnit, VerticalUnit: Integer);
begin
  if not CheckHandle then Exit;
  // There might be a better way to do this...
  if HorizontalUnit > 0 then begin
    while HorizontalUnit > 0 do begin
      SendMessage(HWindow, wm_HScroll, sb_LineRight, 0);
      Dec(HorizontalUnit);
    end;
  end else begin
    while HorizontalUnit < 0 do begin
      SendMessage(HWindow, wm_HScroll, sb_LineLeft, 0);
      Inc(HorizontalUnit);
    end;
  end;
  if VerticalUnit > 0 then begin
    while VerticalUnit > 0 do begin
      SendMessage(HWindow, wm_VScroll, sb_LineDown, 0);
      Dec(VerticalUnit);
    end;
  end else begin
    while VerticalUnit < 0 do begin
      SendMessage(HWindow, wm_VScroll, sb_LineUp, 0);
      Inc(VerticalUnit);
    end;
  end;
end;

function TEdit.Search(StartPos: Integer; AText: string; CaseSensitive: Boolean): Integer;
var
  I: Integer;
  Text: string;
begin
  if not CheckHandle then Exit(0);
  if StartPos=-1 then GetSelection(StartPos, I);
  Text:=GetText;
  if not CaseSensitive then begin
    Text:=LowerCase(Text);
    AText:=LowerCase(AText);
  end;
  I:=Pos(AText, Text);
  if I=0 then Exit(-1);
  Dec(I);
  SetSelection(I, I + Length(AText));
  Result:=I;
end;

function TEdit.SetSelection(StartPos, EndPos: Integer): Boolean;
begin
  if not CheckHandle then Exit(False);
  SendMessage(HWindow, em_SetSel, StartPos, EndPos);
  Result:=True;
end;

procedure TEdit.Undo;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, wm_Undo, 0, 0);
end;

procedure TEdit.SetupWindow;
begin
  inherited SetupWindow;
  SendMessage(HWindow, em_SetLimitText, $7FFFFFFF, 0);
end;

function TEdit.StoreText: Boolean;
begin
  Result:=False;
end;

function TEdit.ProcessParentNotificationMessage(var Msg: TMessage): Boolean;
begin
  case HiWord(Msg.WParam) of
    en_Change:
      Result:=Parent^.ProcessNotification(@Self, cn_Changed);
  end;
end;

{ TListBox }
constructor TListBox.Init(AParent: PWindowsObject; AnId, X, Y, W, H: Integer; Options: TListBoxOptions);
begin
  inherited Init(AParent, AnId, '', X, Y, W, H);
  Attr.Style:=Attr.Style or lbs_NoIntegralHeight or lbs_Notify or ws_VScroll or ws_HScroll;
  if not (loNoBorder in Options) then Attr.ExStyle:=Attr.ExStyle or ws_ex_ClientEdge;
  if loMultiSelect in Options then Attr.Style:=Attr.Style or lbs_ExtendedSel or lbs_MultipleSel;
  if loSortItems in Options then Attr.Style:=Attr.Style or lbs_Sort;
end;

function TListBox.AddString(AString: string): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessageA(HWindow, lb_AddString, 0, LParam(PAnsiChar(AString)));
end;

procedure TListBox.ClearList;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, lb_ResetContent, 0, 0);
end;

function TListBox.DeleteString(Index: Integer): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessage(HWindow, lb_DeleteString, Index, 0);
end;

function TListBox.GetClassName: string;
begin
  Result:='LISTBOX';
end;

function TListBox.GetCount: Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessage(HWindow, lb_GetCount, 0, 0);
end;

function TListBox.GetSelIndex: Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessage(HWindow, lb_GetCurSel, 0, 0);
end;

function TListBox.IsSelected(Index: Integer): Boolean;
begin
  if not CheckHandle then Exit(False);
  Result:=SendMessage(HWindow, lb_GetSel, Index, 0) > 0;
end;

function TListBox.GetSelString: string;
var
  Index: Integer;
begin
  if not CheckHandle then Exit('');
  Index:=GetSelIndex;
  if Index > 0 then
    Result:=GetString(Index)
  else
    Result:='';
end;

function TListBox.GetString(Index: Integer): string;
var
  Len: Integer;
  Buff: PAnsiChar;
begin
  if not CheckHandle then Exit('');
  Len:=SendMessageA(HWindow, lb_GetTextLen, Index, 0);
  if Len <= 0 then Exit('');
  Buff:=GetMem(Len + 1);
  SendMessageA(HWindow, lb_GetText, Index, LParam(Buff));
  Buff[Len]:=#0;
  Result:=Buff;
  FreeMem(Buff);
end;

function TListBox.InsertString(AString: string; Index: Integer): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessageA(HWindow, lb_InsertString, Index, LParam(PAnsiChar(AString)));
end;

function TListBox.SetSelIndex(Index: Integer): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessage(HWindow, lb_SetCurSel, Index, 0);
end;

function TListBox.SetSelected(Index: Integer; Selected: Boolean): Boolean;
begin
  if not CheckHandle then Exit(False);
  if Selected then
    Result:=SendMessage(HWindow, lb_SetSel, Index, 1) <> lb_Err
  else
    Result:=SendMessage(HWindow, lb_SetSel, Index, 0) <> lb_Err;
end;

function TListBox.SetSelString(AString: string; Index: Integer): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessageA(HWindow, lb_SelectString, Index, LParam(PAnsiChar(AString)));
end;

procedure TListBox.SetupWindow;
begin
  inherited SetupWindow;
  if Attr.Title <> '' then SetText(Attr.Title);
end;

function TListBox.StoreText: Boolean;
begin
  Result:=False;
end;

function TListBox.Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer;
var
  Rec: PListBoxTransferRec absolute DataPtr;
  I, Count: Integer;
begin
  Result:=SizeOf(TListBoxTransferRec);
  case TransferFlag of
    tf_GetData: begin
      Count:=GetCount;
      SetLength(Rec^.Strings, Count);
      SetLength(Rec^.Selected, Count);
      Rec^.SelIndex:=GetSelIndex;
      for I:=0 to Count - 1 do begin
        Rec^.Strings[I]:=GetString(I);
        Rec^.Selected[I]:=IsSelected(I);
      end;
    end;
    tf_SetData: begin
      ClearList;
      for I:=0 to High(Rec^.Strings) do
        AddString(Rec^.Strings[I]);
      for I:=0 to High(Rec^.Selected) do
        if Rec^.Selected[I] then
          SetSelected(I, True);
      SetSelIndex(Rec^.SelIndex);
    end;
    tf_SizeData: begin end;
    else Result:=0;
  end;
end;

function TListBox.ProcessParentNotificationMessage(var Msg: TMessage): Boolean;
begin
  case HiWord(Msg.WParam) of
    lbn_SelChange, lbn_SelCancel:
      Result:=Parent^.ProcessNotification(@Self, cn_Changed);
  end;
end;

{ TComboBox }
constructor TComboBox.Init(AParent: PWindowsObject; AnId, X, Y, W, H: Integer; Options: TComboBoxOptions);
begin
  inherited Init(AParent, AnId, '', X, Y, W, H);
  Attr.Style:=Attr.Style or cbs_AutoHScroll or cbs_NoIntegralHeight;
  if coDropDownList in Options then
    Attr.Style:=Attr.Style or cbs_DropDownList
  else
    Attr.Style:=Attr.Style or cbs_DropDown;
  if coSortItems in Options then Attr.Style:=Attr.Style or cbs_Sort;
end;

function TComboBox.AddString(AString: string): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessageA(HWindow, cb_AddString, 0, LParam(PAnsiChar(AString)));
end;

procedure TComboBox.ClearList;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, cb_ResetContent, 0, 0);
end;

function TComboBox.DeleteString(Index: Integer): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessage(HWindow, cb_DeleteString, Index, 0);
end;

function TComboBox.GetClassName: string;
begin
  Result:='COMBOBOX';
end;

function TComboBox.GetCount: Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessage(HWindow, cb_GetCount, 0, 0);
end;

function TComboBox.GetSelIndex: Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessage(HWindow, cb_GetCurSel, 0, 0);
end;

function TComboBox.GetSelString: string;
var
  Index: Integer;
begin
  if not CheckHandle then Exit('');
  Index:=GetSelIndex;
  if Index > 0 then
    Result:=GetString(Index)
  else
    Result:='';
end;

function TComboBox.GetString(Index: Integer): string;
var
  Len: Integer;
  Buff: PAnsiChar;
begin
  if not CheckHandle then Exit('');
  Len:=SendMessageA(HWindow, cb_GetLBTextLen, Index, 0);
  if Len <= 0 then Exit('');
  Buff:=GetMem(Len + 1);
  SendMessageA(HWindow, cb_GetLBText, Index, LParam(Buff));
  Buff[Len]:=#0;
  Result:=Buff;
  FreeMem(Buff);
end;

function TComboBox.InsertString(AString: string; Index: Integer): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessageA(HWindow, cb_InsertString, Index, LParam(PAnsiChar(AString)));
end;

function TComboBox.SetSelIndex(Index: Integer): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessage(HWindow, cb_SetCurSel, Index, 0);
end;

function TComboBox.SetSelString(AString: string; Index: Integer): Integer;
begin
  if not CheckHandle then Exit(-1);
  Result:=SendMessageA(HWindow, cb_SelectString, Index, LParam(PAnsiChar(AString)));
end;

procedure TComboBox.HideList;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, cb_ShowDropDown, 0, 0);
end;

procedure TComboBox.ShowList;
begin
  if not CheckHandle then Exit;
  SendMessage(HWindow, cb_ShowDropDown, 1, 0);
end;

procedure TComboBox.SetupWindow;
begin
  SendMessage(HWindow, cb_LimitText, 0, 0);
  inherited SetupWindow;
end;

function TComboBox.Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer;
var
  Rec: PComboBoxTransferRec absolute DataPtr;
  I, Count: Integer;
begin
  Result:=SizeOf(TListBoxTransferRec);
  case TransferFlag of
    tf_GetData: begin
      Count:=GetCount;
      SetLength(Rec^.Strings, Count);
      Rec^.SelIndex:=GetSelIndex;
      for I:=0 to Count - 1 do
        Rec^.Strings[I]:=GetString(I);
      Rec^.Text:=GetText;
    end;
    tf_SetData: begin
      ClearList;
      for I:=0 to High(Rec^.Strings) do
        AddString(Rec^.Strings[I]);
      SetSelIndex(Rec^.SelIndex);
      SetText(Rec^.Text);
    end;
    tf_SizeData: begin end;
    else Result:=0;
  end;
end;

function TComboBox.ProcessParentNotificationMessage(var Msg: TMessage): Boolean;
begin
  case HiWord(Msg.WParam) of
    cbn_SelChange:
      Result:=Parent^.ProcessNotification(@Self, cn_Changed);
  end;
end;

{ TGroupBox }
constructor TGroupBox.Init(AParent: PWindowsObject; AnId: Integer; ATitle: string; X, Y, W, H: Integer);
begin
  inherited Init(AParent, AnId, ATitle, X, Y, W, H);
  NotifyParent:=True;
  Attr.Style:=Attr.Style or bs_GroupBox;
  Attr.Style:=Attr.Style and (not ws_TabStop);
end;

constructor TGroupBox.Load(var S: TStream);
begin
  inherited Load(S);
  S.Read(NotifyParent, 1);
end;

function TGroupBox.GetClassName: string;
begin
  Result:='BUTTON';
end;

procedure TGroupBox.SelectionChanged(ControlId: Integer);
var
  Child: PWindowsObject;
begin
  Child:=ChildWithId(ControlId);
  if Assigned(Parent) and Assigned(Child) then
    Parent^.ProcessNotification(Child, cn_Changed);
end;

procedure TGroupBox.SetupWindow;
begin
  inherited SetupWindow;
  if not ButtonSubclassInstalled then begin
    ButtonSubclassInstalled:=True;
    {$IFDEF CPUAMD64}
    OriginalButtonProc:=WndProc(GetClassLongPtr(HWindow, GCLP_WNDPROC));
    SetClassLongPtr(HWindow, GCLP_WNDPROC, LONG_PTR(@ReplacementButtonProc));
    {$ELSE}
    OriginalButtonProc:=WndProc(GetClassLong(HWindow, GCL_WNDPROC));
    SetClassLong(HWindow, GCL_WNDPROC, LONG(@ReplacementButtonProc));
    {$ENDIF}
    Destroy;
    Create;
  end;
end;

procedure TGroupBox.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(NotifyParent, 1);
end;

function TGroupBox.ProcessNotification(Child: PWindowsObject; NotifyCode: Integer): Boolean;
begin
  if NotifyParent and Assigned(Parent) then
    Result:=Parent^.ProcessNotification(Child, NotifyCode)
  else
    Result:=False;
end;

{ TCheckBox }
constructor TCheckBox.Init(AParent: PWindowsObject; AnId: Integer; ATitle: string; X, Y, W, H: Integer; Options: TCheckBoxOptions);
begin
  inherited Init(AParent, AnId, ATitle, X, Y, W, H);
  if co3State in Options then begin
    if coAutoCheck in Options then
      Attr.Style:=Attr.Style or bs_Auto3State
    else
      Attr.Style:=Attr.Style or bs_3State;
  end else begin
    if coAutoCheck in Options then
      Attr.Style:=Attr.Style or bs_AutoCheckBox
    else
      Attr.Style:=Attr.Style or bs_CheckBox;
  end;
end;

procedure TCheckBox.Check;
begin
  SetCheck(csChecked);
end;

function TCheckBox.GetClassName: string;
begin
  Result:='BUTTON';
end;

function TCheckBox.GetCheck: TCheckBoxState;
begin
  if not CheckHandle then Exit(csUnchecked);
  case SendMessage(HWindow, bm_GetCheck, 0, 0) of
    bst_Checked: Result:=csChecked;
    bst_Indeterminate: Result:=csGrayed;
    else Result:=csUnchecked;
  end;
end;

procedure TCheckBox.SetCheck(CheckFlag: TCheckBoxState);
begin
  if not CheckHandle then Exit;
  case CheckFlag of
    csUnchecked: SendMessage(HWindow, bm_SetCheck, bst_Unchecked, 0);
    csChecked: SendMessage(HWindow, bm_SetCheck, bst_Checked, 0);
    csGrayed: SendMessage(HWindow, bm_SetCheck, bst_Indeterminate, 0);
  end;
end;

procedure TCheckBox.Toggle;
var
  Style: DWORD;
begin
  if not CheckHandle then Exit;
  Style:=GetWindowLong(HWindow, GWL_STYLE);
  if ((Style and bs_3State)=bs_3State) or
     ((Style and bs_Auto3State)=bs_Auto3State) then begin
    case GetCheck of
      csUnchecked: SetCheck(csChecked);
      csChecked: SetCheck(csGrayed);
      csGrayed: SetCheck(csUnchecked);
    end;
  end else begin
    if GetCheck=csChecked then
      SetCheck(csUnchecked)
    else
      SetCheck(csChecked);
  end;
end;

function TCheckBox.Transfer(DataPtr: Pointer; TransferFlag: Integer): Integer;
begin
  Result:=Sizeof(TCheckBoxState);
  case TransferFlag of
    tf_GetData: PCheckBoxState(DataPtr)^:=GetCheck;
    tf_SetData: SetCheck(PCheckBoxState(DataPtr)^);
    tf_SizeData: begin end;
    else Result:=0;
  end;
end;

procedure TCheckBox.Uncheck;
begin
  SetCheck(csUnchecked);
end;

function TCheckBox.ProcessParentNotificationMessage(var Msg: TMessage): Boolean;
begin
  Result:=False;
  case HiWord(Msg.WParam) of
    Windows.bn_Clicked:
      Result:=Parent^.ProcessNotification(@Self, cn_Clicked);
  end;
end;

{ TRadioButton }
constructor TRadioButton.Init(AParent: PWindowsObject; AnId: Integer; ATitle: string; X, Y, W, H: Integer);
begin
  inherited Init(AParent, AnId, ATitle, X, Y, W, H, []);
  Attr.Style:=Attr.Style and (not bs_CheckBox);
  Attr.Style:=Attr.Style or bs_RadioButton;
end;

procedure TRadioButton.SetCheck(CheckFlag: TCheckBoxState);
var
  Child: PWindowsObject;
begin
  // Normalize the value
  if CheckFlag <> csChecked then CheckFlag:=csUnchecked;
  // Make sure no other radio button in the parent window
  // is checked
  if Assigned(Parent) and (CheckFlag=csChecked) then begin
    Child:=Parent^.ChildList;
    while Assigned(Child) do begin
      if PPointer(@Self)^=PPointer(Child)^ then
        PRadioButton(Child)^.Uncheck;
      Child:=Child^.Next;
    end;
  end;
  inherited SetCheck(CheckFlag);
end;

function TRadioButton.ProcessParentNotificationMessage(var Msg: TMessage): Boolean;
begin
  Result:=False;
  case HiWord(Msg.WParam) of
    Windows.bn_Clicked: begin
      Check;
      Result:=Parent^.ProcessNotification(@Self, cn_Clicked);
    end;
  end;
end;

{ TApplication }
constructor TApplication.Init(AName: string);
begin
  inherited Init;
  KBHandlerWnd:=nil;
  MainWindow:=nil;
  Name:=AName;
  Status:=0;
  Application:=@Self;
  InitApplication;
  InitInstance;
end;

destructor TApplication.Done;
begin
  if Assigned(MainWindow) then Dispose(MainWindow, Done);
  inherited Done;
end;

function TApplication.CanClose: Boolean;
begin
  if Assigned(MainWindow) then
    Result:=MainWindow^.CanClose
  else
    Result:=True;
end;

procedure TApplication.Error(ErrorCode: Integer);
var
  Msg: AnsiString;
  Handle: THandle;
  S: string;
begin
  Status:=ErrorCode;
  Str(ErrorCode, S);
  Msg:='Application error #' + S + '. Do you want to continue? Press No to terminate the program.';
  Handle:=0;
  if Assigned(MainWindow) then Handle:=MainWindow^.HWindow;
  if MessageBoxA(Handle, PAnsiChar(Msg), 'Error', MB_ICONERROR or MB_YESNO)=IDNO then Halt;
end;

function TApplication.ExecDialog(ADialog: PWindowsObject): Integer;
begin
  ADialog:=ValidWindow(ADialog);
  if Assigned(ADialog) then begin
    // TODO
  end;
  Result:=0;
end;

procedure TApplication.ShowMessage(Text: string);
var
  Wnd: THandle;
begin
  if Assigned(MainWindow) then
    Wnd:=MainWindow^.HWindow
  else
    Wnd:=0;
  MessageBoxA(Wnd, PAnsiChar(Text), PAnsiChar(Name), MB_OK);
end;

procedure TApplication.InitApplication;
begin
end;

procedure TApplication.InitInstance;
begin
  InitMainWindow;
  MainWindow:=MakeWindow(MainWindow);
  if not Assigned(MainWindow) then begin
    Error(em_InvalidMainWindow);
    Exit;
  end;
  MainWindow^.Show(sw_Show);
end;

procedure TApplication.InitMainWindow;
begin
  MainWindow:=MakeWindow(New(PWindow, Init(nil, 'Fowl Window')));
end;

function TApplication.MakeWindow(AWindowsObject: PWindowsObject): PWindowsObject;
begin
  if not Assigned(AWindowsObject) then Exit(nil);
  if not Assigned(ValidWindow(AWindowsObject)) then begin
    Dispose(AWindowsObject, Done);
    Exit(nil);
  end;
  if not AWindowsObject^.Create then begin
    Dispose(AWindowsObject, Done);
    Exit(nil);
  end;
  Result:=AWindowsObject;
end;

procedure TApplication.MessageLoop;
var
  Msg: TMsg;
  Mess: TMessage;
begin
  while GetMessage(Msg, 0, 0, 0) do begin
    Mess.Receiver:=Msg.HWnd;
    Mess.Message:=Msg.Message;
    Mess.WParam:=Msg.WParam;
    Mess.LParam:=Msg.LParam;
    Mess.Result:=0;
    if ProcessAppMsg(Mess) then begin
      Msg.HWnd:=Mess.Receiver;
      Msg.Message:=Msg.Message;
      Msg.WParam:=Mess.WParam;
      Msg.LParam:=Mess.LParam;
    end;
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;
end;

function TApplication.ProcessAccels(var Message: TMessage): Boolean;
begin
  Result:=False;
end;

function TApplication.ProcessAppMsg(var Message: TMessage): Boolean;
begin
  if ProcessDlgMsg(Message) then Exit(True);
  if ProcessMDIAccels(Message) then Exit(True);
  if ProcessAccels(Message) then Exit(True);
  Result:=False;
end;

function TApplication.ProcessDlgMsg(var Message: TMessage): Boolean;
begin
  Result:=False;
end;

function TApplication.ProcessMDIAccels(var Message: TMessage): Boolean;
begin
  Result:=False;
end;

procedure TApplication.Run;
begin
  if Status=0 then MessageLoop;
end;

procedure TApplication.Terminate;
begin
  PostQuitMessage(0);
end;

procedure TApplication.SetKBHandler(AWindowsObject: PWindowsObject);
begin
  KBHandlerWnd:=AWindowsObject;
end;

function TApplication.ValidWindow(AWindowsObject: PWindowsObject): PWindowsObject;
begin
  if not Assigned(AWindowsObject) then Exit(nil);
  if AWindowsObject^.Status <> 0 then Exit(nil);
  Result:=AWindowsObject;
end;

const
  { Stream records }
  RCollection: TStreamRec = (
    ObjType: 1;
    VmtLink: Ofs(TypeOf(TCollection)^);
    Load: @TCollection.Load;
    Store: @TCollection.Store;
    Next: nil);
  RMenuItem: TStreamRec = (
    ObjType: 2;
    VmtLink: Ofs(TypeOf(TMenuItem)^);
    Load: @TMenuItem.Load;
    Store: @TMenuItem.Store;
    Next: nil);
  RMenu: TStreamRec = (
    ObjType: 3;
    VmtLink: Ofs(TypeOf(TMenu)^);
    Load: @TMenu.Load;
    Store: @TMenu.Store;
    Next: nil);
  RWindowsObject: TStreamRec = (
    ObjType: 4;
    VmtLink: Ofs(TypeOf(TWindowsObject)^);
    Load: @TWindowsObject.Load;
    Store: @TWindowsObject.Store;
    Next: nil);
  RWindow: TStreamRec = (
    ObjType: 5;
    VmtLink: Ofs(TypeOf(TWindow)^);
    Load: @TWindow.Load;
    Store: @TWindow.Store;
    Next: nil);
  RButton: TStreamRec = (
    ObjType: 6;
    VmtLink: Ofs(TypeOf(TButton)^);
    Load: @TButton.Load;
    Store: @TButton.Store;
    Next: nil);
  RScrollBar: TStreamRec = (
    ObjType: 7;
    VmtLink: Ofs(TypeOf(TScrollBar)^);
    Load: @TScrollBar.Load;
    Store: @TScrollBar.Store;
    Next: nil);
  RStatic: TStreamRec = (
    ObjType: 8;
    VmtLink: Ofs(TypeOf(TStatic)^);
    Load: @TStatic.Load;
    Store: @TStatic.Store;
    Next: nil);
  REdit: TStreamRec = (
    ObjType: 9;
    VmtLink: Ofs(TypeOf(TEdit)^);
    Load: @TEdit.Load;
    Store: @TEdit.Store;
    Next: nil);
  RListBox: TStreamRec = (
    ObjType: 10;
    VmtLink: Ofs(TypeOf(TListBox)^);
    Load: @TListBox.Load;
    Store: @TListBox.Store;
    Next: nil);
  RComboBox: TStreamRec = (
    ObjType: 11;
    VmtLink: Ofs(TypeOf(TComboBox)^);
    Load: @TComboBox.Load;
    Store: @TComboBox.Store;
    Next: nil);
  RGroupBox: TStreamRec = (
    ObjType: 12;
    VmtLink: Ofs(TypeOf(TGroupBox)^);
    Load: @TGroupBox.Load;
    Store: @TGroupBox.Store;
    Next: nil);
  RCheckBox: TStreamRec = (
    ObjType: 13;
    VmtLink: Ofs(TypeOf(TCheckBox)^);
    Load: @TCheckBox.Load;
    Store: @TCheckBox.Store;
    Next: nil);
  RRadioButton: TStreamRec = (
    ObjType: 14;
    VmtLink: Ofs(TypeOf(TRadioButton)^);
    Load: @TRadioButton.Load;
    Store: @TRadioButton.Store;
    Next: nil);
initialization
  StreamRecHead:=nil;
  StreamError:=@DefaultStreamError;
  StreamVersion:=FowlVersion;
  RegisterType(RCollection);
  RegisterType(RMenuItem);
  RegisterType(RMenu);
  RegisterType(RWindowsObject);
  RegisterType(RWindow);
  RegisterType(RButton);
  RegisterType(RScrollBar);
  RegisterType(RStatic);
  RegisterType(REdit);
  RegisterType(RListBox);
  RegisterType(RComboBox);
  RegisterType(RGroupBox);
  RegisterType(RCheckBox);
  RegisterType(RRadioButton);
end.
